from __future__ import annotations

import json
import os
import random
import string
import warnings
from dataclasses import asdict
from typing import Any, Dict, List, Literal, MutableMapping, Sequence, Union, cast

import networkx as nx
import numpy as np
import pandas as pd
from IPython.core.display import HTML, display
from nanoid import generate

from retentioneering import RETE_CONFIG, __version__
from retentioneering.backend import ServerManager
from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.edgelist import Edgelist
from retentioneering.eventstream.types import EventstreamType
from retentioneering.nodelist import Nodelist
from retentioneering.nodelist.nodelist import (
    IS_AGGREGATED_COL,
    IS_DISABLED_COL,
    IS_GROUP_COL,
    IS_PINNED_COL,
    NAME_COL,
    PARENT_ID_COL,
)
from retentioneering.templates.transition_graph import TransitionGraphRenderer

from .types import (
    Column,
    EdgeId,
    EdgeItem,
    EdgesCustomColors,
    NodeId,
    NodeItem,
    NodePoint,
    NodesCustomColors,
    NodesPoints,
    NodeTarget,
    NodeTargetId,
    Normalization,
    NormType,
    RecalculationEdgeItem,
    RecalculationNodeItem,
    RecalculationResult,
    SerializedEdgesState,
    SerializedNodesState,
    SerializedSettings,
    SerializedState,
    Settings,
    StateChanges,
    TargetToNodesMap,
    ThresholdValueMap,
    ThresholdWithFallback,
    Tracker,
    Weight,
)

SESSION_ID_COL = "session_id"


class TransitionGraph:
    """
    A class that holds methods for transition graph visualization.

    Parameters
    ----------
    eventstream: EventstreamType
        Source eventstream.


    See Also
    --------
    .Eventstream.transition_graph : Call TransitionGraph tool as an eventstream method.
    .Eventstream.transition_matrix : Matrix representation of transition graph.
    .EventstreamSchema : Schema of eventstream columns, that could be used as weights.
    .TransitionGraph.plot : Interactive transition graph visualization.


    Notes
    -----
    See :doc:`transition graph user guide</user_guides/transition_graph>` for the details.

    """

    DEFAULT_GRAPH_URL = "https://static.server.retentioneering.com/package/@rete/transition-graph/version/4/dist/transition-graph.umd.js"

    _weights: MutableMapping[str, str] | None

    _nodes_norm_type: NormType
    _edges_norm_type: NormType

    _nodes_threshold: ThresholdValueMap = {}
    _edges_threshold: ThresholdValueMap = {}

    _recalculation_result: EventstreamType

    _allowed_targets: list[NodeTarget] = [
        NodeTarget(
            id="positive",
            name="Positive",
            ignoreThreshold=True,
            edgeDirection="in",
            position="top-right",
        ),
        NodeTarget(
            id="negative",
            name="Negative",
            ignoreThreshold=True,
            edgeDirection="in",
            position="bottom-right",
        ),
        NodeTarget(
            id="source",
            name="Source",
            ignoreThreshold=False,
            edgeDirection="both",
            position="top-left",
        ),
    ]

    _allowed_normalizations: list[Normalization] = [
        Normalization(id="none", name="none", type="absolute"),
        Normalization(id="full", name="full", type="relative"),
        Normalization(id="node", name="node", type="relative"),
    ]

    @property
    def graph_url(self) -> str:
        env_url: str = os.getenv("RETE_TRANSITION_GRAPH_URL", "")
        return env_url if env_url else self.DEFAULT_GRAPH_URL

    @property
    def nodes_thresholds(self) -> ThresholdValueMap:
        return self._nodes_threshold

    @nodes_thresholds.setter
    def nodes_thresholds(self, value: ThresholdValueMap) -> None:
        if self._check_thresholds_for_norm_type(value=value, norm_type=self.nodes_norm_type):
            self._nodes_threshold = value

    @property
    def edges_thresholds(self) -> ThresholdValueMap:
        return self._edges_threshold

    @edges_thresholds.setter
    def edges_thresholds(self, value: ThresholdValueMap) -> None:
        if self._check_thresholds_for_norm_type(value=value, norm_type=self.edges_norm_type):
            self._edges_threshold = value

    @staticmethod
    def _check_thresholds_for_norm_type(value: ThresholdValueMap, norm_type: NormType) -> bool:
        values: List[Union[int, float]] = [val for key in value.values() for val in key.values()]  # type: ignore

        if norm_type is None:
            if not all(map(lambda x: x is None or x >= 0, values)):  # type: ignore
                raise ValueError(f"For normalization type {norm_type} all thresholds must be positive or None")
        else:
            if not all(map(lambda x: x is None or 0 <= x <= 1, values)):  # type: ignore
                raise ValueError(f"For normalization type {norm_type} all thresholds must be between 0 and 1 or None")

        return True

    @time_performance(
        scope="transition_graph",
        event_name="init",
    )
    def __init__(
        self,
        eventstream: EventstreamType,  # graph: dict,  # preprocessed graph
    ) -> None:
        from retentioneering.eventstream.eventstream import Eventstream

        sm = ServerManager()
        self.sm = sm

        self.server = sm.create_server()
        self.bridge_id = sm.widget.bridge_id if sm.widget else ""

        self.server.register_action("recalculate", lambda n: self._on_recalc_request(n))

        self.server.register_action("test", lambda n: self._on_test_request(n))

        self.eventstream: Eventstream = eventstream  # type: ignore

        self.event_col = self.eventstream.schema.event_name
        self.event_time_col = self.eventstream.schema.event_timestamp
        self.user_col = self.eventstream.schema.user_id

        self.weight_cols = self._define_weight_cols(None)

        self.nodes_weight_col = self.eventstream.schema.user_id
        self.edges_weight_col = self.eventstream.schema.user_id

        # TODO: WIP Layout experiments
        self.spring_layout_config = {"k": 0.1, "iterations": 300, "nx_threshold": 1e-4}

        self.render: TransitionGraphRenderer = TransitionGraphRenderer()

        self._recalculation_result = eventstream

        self.targets: TargetToNodesMap = {}

        self.nodes_weights_range: ThresholdValueMap = {}
        self.edges_weights_range: ThresholdValueMap = {}

        self.nodes_custom_colors: NodesCustomColors = {}
        self.edges_custom_colors: EdgesCustomColors = {}

        self.settings = Settings(
            show_weights=RETE_CONFIG.transition_graph.show_weights,
            show_percents=RETE_CONFIG.transition_graph.show_percents,
            show_nodes_names=RETE_CONFIG.transition_graph.show_nodes_names,
            show_all_edges_for_targets=RETE_CONFIG.transition_graph.show_all_edges_for_targets,
            show_nodes_without_links=RETE_CONFIG.transition_graph.show_nodes_without_links,
            show_edge_info_on_hover=RETE_CONFIG.transition_graph.show_edge_info_on_hover,
            open_sidebar_by_default=RETE_CONFIG.transition_graph.open_sidebar_by_default,
        )

        self.width: str | int | float = RETE_CONFIG.transition_graph.width
        self.height: str | int | float = RETE_CONFIG.transition_graph.height

        self.nodes_points: NodesPoints = {}

        self._state_changes = StateChanges()
        self._recalculation_changes = StateChanges()

    @property
    @time_performance(
        scope="transition_graph",
        event_name="recalculation_result",
    )
    def recalculation_result(self) -> EventstreamType:
        """
        Export an eventstream after GUI actions that affect eventstream.

        Returns
        -------
        EventstreamType
            The modified event stream.

        Notes
        -----
        Renaming groups, nodes, and nested nodes in the GUI will not affect the resulting eventstream.
        The default group and node names will be returned.
        """
        return self._recalculation_result

    def _define_weight_cols(self, custom_weight_cols: list[str] | None) -> list[str]:
        weight_cols = [
            self.eventstream.schema.event_id,
            self.eventstream.schema.user_id,
        ]
        if SESSION_ID_COL in self.eventstream.schema.custom_cols:
            weight_cols.append(SESSION_ID_COL)
        if isinstance(custom_weight_cols, list):
            for col in custom_weight_cols:
                if col not in weight_cols:
                    if col not in self.eventstream.schema.custom_cols:
                        raise ValueError(f"Custom weights column {col} not found in eventstream schema")
                    else:
                        weight_cols.append(col)
        return weight_cols

    @property
    def weights(self) -> MutableMapping[str, str] | None:
        return self._weights

    @weights.setter
    def weights(self, value: MutableMapping[str, str] | None) -> None:
        available_cols = self.__get_nodelist_cols()

        if value and ("edges" not in value or "nodes" not in value):
            raise ValueError("Allowed only: %s" % {"edges": "col_name", "nodes": "col_name"})

        if value and (value["edges"] not in available_cols or value["nodes"] not in available_cols):
            raise ValueError("Allowed only: %s" % {"edges": "col_name", "nodes": "col_name"})

        self._weights = value

    @property
    def edges_norm_type(self) -> NormType:
        return self._edges_norm_type

    @edges_norm_type.setter
    def edges_norm_type(self, edges_norm_type: NormType) -> None:
        allowed_edges_norm_types: list[str | None] = [None, "full", "node"]
        if edges_norm_type in allowed_edges_norm_types:
            self._edges_norm_type = edges_norm_type
        else:
            raise ValueError("Norm type should be one of: %s" % allowed_edges_norm_types)

    @property
    def nodes_norm_type(self) -> NormType:
        return self._nodes_norm_type

    @nodes_norm_type.setter
    def nodes_norm_type(self, nodes_norm_type: NormType) -> None:
        if nodes_norm_type is not None:
            warnings.warn(f"Currently nodes_norm_type allowed to be None only")
        self._nodes_norm_type = None

    @staticmethod
    def _extract_nodes_points_from_serialized_state(serialized_state: SerializedState) -> NodesPoints:
        result: NodesPoints = {}

        for node in serialized_state.nodes["items"]:
            x = node.get("x")
            y = node.get("y")
            if x is not None and y is not None:
                result[node["id"]] = NodePoint(x=x, y=y)

        return result

    def make_nodelist_from_serialized_state(self, serialized_state: SerializedState) -> pd.DataFrame:
        nodelist_data: List[Dict[str, Any]] = []
        weight_cols = self.weight_cols

        for node in serialized_state.nodes["items"]:
            new_node: Dict[str, Any] = {self.event_col: node["id"]}
            # TODO: WIP Refactor me
            visibility_rule = node.get("visibilityRule", None)
            new_node[IS_DISABLED_COL] = visibility_rule == "disabled"
            new_node[IS_PINNED_COL] = visibility_rule == "pinned"
            new_node[PARENT_ID_COL] = node.get("parentNodeId", None)
            new_node[IS_GROUP_COL] = node.get("isGroup", False)
            new_node[IS_AGGREGATED_COL] = node.get("isAggregated", False)
            name = node.get("name")
            new_node[NAME_COL] = name if name != new_node[self.event_col] else new_node[self.event_col]

            weights: Dict[str, float] = node.get("weight", {})

            for weight_col in weight_cols:
                weight = weights.get(weight_col, None)
                if weight is not None:
                    new_node[weight_col] = weight

            nodelist_data.append(new_node)

        return pd.DataFrame(nodelist_data)

    def _on_test_request(self, payload: dict[str, Any]) -> dict[str, Any]:
        return {"status": "ok", "message": "Test request received"}

    def _on_recalc_request(self, payload: dict[str, Any]) -> dict[str, Any]:
        try:
            t = json.loads(payload["recalcjson"])
            serialized_state: SerializedState = self.__build_serialized_state_from_unknown_data(t)
        except Exception:
            raise Exception("Invalid recalculate data")

        try:
            self.nodes_points = self._extract_nodes_points_from_serialized_state(serialized_state)

            self._state_changes = serialized_state.stateChanges
            self._recalculation_changes = serialized_state.recalculationChanges

            self.nodes_thresholds = serialized_state.stateChanges.nodesThreshold
            self.edges_thresholds = serialized_state.stateChanges.edgesThreshold

            new_nodelist = self.make_nodelist_from_serialized_state(serialized_state)
            self._recalculation_result = self.__apply_nodelist(new_nodelist)

            nodes_items = self._prepare_nodes(nodelist=self.nodelist.nodelist_df)
            edgelist = self.edgelist.get_filtered_edges()
            edgelist["type"] = "suit"

            recalculation_response = self._build_recalculation_response(
                serialized_state=serialized_state,
                nodes=nodes_items,
                edges=self._prepare_edges(edgelist=edgelist, nodes_items=nodes_items),
            )

            # noinspection PyTypeChecker
            # see https://youtrack.jetbrains.com/issue/PY-76059/Incorrect-Type-warning-with-asdict-and-Dataclass
            return asdict(recalculation_response)
        except Exception as err:
            raise ValueError("error! %s" % err)

    def _build_recalculation_response(
        self, serialized_state: SerializedState, nodes: list[NodeItem], edges: list[EdgeItem]
    ) -> RecalculationResult:
        renamed_nodes = {}
        for node_item in serialized_state.nodes["items"]:
            renamed_nodes[node_item["id"]] = node_item["name"]

        node_edge_map: dict[str, EdgeId] = {}
        node_id_name_mapping: dict[NodeId, str] = {node["id"]: node["name"] for node in serialized_state.nodes["items"]}
        node_name_id_mapping: dict[NodeId, str] = {node["name"]: node["id"] for node in serialized_state.nodes["items"]}
        for edge in serialized_state.edges["items"]:
            node_edge_map[
                f'{[node_id_name_mapping[edge["sourceNodeId"]]]}, {node_id_name_mapping[edge["targetNodeId"]]}'
            ] = edge["id"]

        response_nodes: dict[NodeId, RecalculationNodeItem] = {
            node_name_id_mapping.get(node["id"], node["id"]): RecalculationNodeItem(
                id=node_name_id_mapping.get(node["id"], node["id"]),
                size=node["size"],
                weight=node["weight"],
                targetId=node.get("targetId"),
            )
            for node in nodes
        }
        self._response_nodes = response_nodes

        response_edges: list[RecalculationEdgeItem] = []
        for edge in edges:
            response_edges.append(
                RecalculationEdgeItem(
                    id=node_edge_map.get(f'{[edge["sourceNodeId"], edge["targetNodeId"]]}', edge["id"]),
                    sourceNodeId=node_name_id_mapping[renamed_nodes.get(edge["sourceNodeId"], edge["sourceNodeId"])],
                    targetNodeId=node_name_id_mapping[renamed_nodes.get(edge["targetNodeId"], edge["targetNodeId"])],
                    size=edge["size"],
                    weight=edge["weight"],
                )
            )

        return RecalculationResult(
            nodes=response_nodes,
            edges=response_edges,
            nodesThresholds=self.nodes_thresholds,
            edgesThresholds=self.edges_thresholds,
            nodesThresholdsMinMax=self.nodelist.get_min_max(),
            edgesThresholdsMinMax=self.edgelist.get_min_max(),
        )

    def __get_nodelist_cols(self) -> list[str]:
        default_col = self.nodelist_default_col
        custom_cols = self.weight_cols
        return list([default_col]) + list(custom_cols)

    # TODO: WIP Layout experiments
    def _calc_layout(self, edgelist: pd.DataFrame, width: int, height: int) -> MutableMapping[str, Sequence[float]]:
        directed_graph = nx.DiGraph()
        source_col = edgelist.columns[0]
        target_col = edgelist.columns[1]
        weight_col = edgelist.columns[2]

        directed_graph.add_weighted_edges_from(edgelist.loc[:, [source_col, target_col, weight_col]].values)

        pos = nx.layout.spring_layout(
            directed_graph,
            k=self.spring_layout_config["k"],
            iterations=self.spring_layout_config["iterations"],
            threshold=self.spring_layout_config["nx_threshold"],
            seed=0,
        )

        all_x_coords: list[float] = []
        all_y_coords: list[float] = []

        for j in pos.values():
            all_x_coords.append(j[0])
            all_y_coords.append(j[1])

        min_x = min(all_x_coords)
        min_y = min(all_y_coords)
        max_x = max(all_x_coords)
        max_y = max(all_y_coords)

        return {
            i: [
                (j[0] - min_x) / (max_x - min_x) * (width - 150) + 75,
                (j[1] - min_y) / (max_y - min_y) * (height - 100) + 50,
            ]
            for i, j in pos.items()
        }

    def _prepare_nodes(
        self,
        nodelist: pd.DataFrame,
        mapped_node_targets: dict[NodeId, NodeTargetId] | None = None,
        nodes_custom_colors: NodesCustomColors | None = None,
    ) -> list[NodeItem]:
        node_names: set[str] = set(nodelist[self.event_col])
        nodes_custom_colors = nodes_custom_colors if nodes_custom_colors else {}

        cols = self.__get_nodelist_cols()
        nodes_set: list[NodeItem] = []

        for idx, node_name in enumerate(node_names):
            row = nodelist.loc[nodelist[self.event_col] == node_name]
            degree = {}
            weight = {}
            size = {}
            for weight_col in cols:
                max_degree = cast(float, nodelist[weight_col].max())
                r = row[weight_col]
                r = r.tolist()
                value = r[0]
                curr_degree = {
                    "degree": (abs(value)) / abs(max_degree) * 30 + 4,
                    "source": value,
                }
                degree[weight_col] = curr_degree
                size[weight_col] = curr_degree["degree"]
                weight[weight_col] = curr_degree["source"]

            nodelist_row_dict: dict = nodelist[nodelist[self.event_col] == node_name].iloc[0].to_dict()

            node = NodeItem(
                id=node_name,
                size=size,
                weight=weight,
                name=nodelist_row_dict[NAME_COL],
                isAggregated=nodelist_row_dict[IS_AGGREGATED_COL],
                isGroup=nodelist_row_dict[IS_GROUP_COL],
                children=nodelist.loc[nodelist[PARENT_ID_COL] == node_name, self.event_col].tolist(),
            )

            if mapped_node_targets is not None:
                target_node_id = mapped_node_targets.get(node_name)
                if target_node_id is not None:
                    node["targetId"] = target_node_id

            parent_node_id = nodelist_row_dict[PARENT_ID_COL]
            if parent_node_id is not None:
                node["parentNodeId"] = parent_node_id

            visibility_rule = None
            if nodelist_row_dict[IS_PINNED_COL]:
                visibility_rule = "pinned"
            elif nodelist_row_dict[IS_DISABLED_COL]:
                visibility_rule = "disabled"

            if visibility_rule is not None:
                node["visibilityRule"] = visibility_rule

            custom_color = nodes_custom_colors.get(node_name)
            if custom_color is not None:
                node["customColor"] = custom_color

            node_point = self.nodes_points.get(node_name)
            if node_point is not None:
                node["x"] = node_point["x"]
                node["y"] = node_point["y"]

            nodes_set.append(node)

        return nodes_set

    def _prepare_edges(
        self,
        edgelist: pd.DataFrame,
        nodes_items: list[NodeItem],
        edges_custom_colors: EdgesCustomColors | None = None,
    ) -> list[EdgeItem]:
        default_col = self.nodelist_default_col
        source_col = edgelist.columns[0]
        target_col = edgelist.columns[1]
        weight_col = edgelist.columns[2]
        custom_cols = self.weight_cols
        edges: list[EdgeItem] = []
        edges_custom_colors = edges_custom_colors if edges_custom_colors else {}

        edgelist["weight_norm"] = edgelist[weight_col] / edgelist[weight_col].abs().max()
        for _, row in edgelist.iterrows():
            default_col_weight: Weight = {
                "weight_norm": row.weight_norm,
                "weight": cast(float, row[weight_col]),
            }
            weights = {
                default_col: default_col_weight,
            }
            for custom_weight_col in custom_cols:
                weight = cast(float, row[custom_weight_col])
                max_weight = cast(float, edgelist[custom_weight_col].abs().max())
                weight_norm = weight / max_weight
                col_weight: Weight = {
                    "weight_norm": weight_norm,
                    "weight": weight,
                }
                weights[custom_weight_col] = col_weight

            source_node_name = str(row[source_col])
            target_node_name = str(row[target_col])

            custom_color = edges_custom_colors.get((source_node_name, target_node_name))

            # list comprehension faster than filter
            source_node = [node for node in nodes_items if node["id"] == source_node_name][0]
            target_node = [node for node in nodes_items if node["id"] == target_node_name][0]

            if source_node is not None and target_node is not None:  # type: ignore
                edge_item = EdgeItem(
                    id=generate(),
                    sourceNodeId=source_node_name,
                    targetNodeId=target_node_name,
                    weight={col_name: weight["weight"] for col_name, weight in weights.items()},
                    size={col_name: weight["weight_norm"] for col_name, weight in weights.items()},
                    aggregatedEdges=[],
                )

                if custom_color is not None:
                    edge_item["customColor"] = custom_color

                edges.append(edge_item)

        return edges

    @staticmethod
    def _to_json(data: Any) -> str:
        def convert_numpy(obj: Any) -> int:
            if isinstance(obj, np.integer):
                return int(obj)
            raise TypeError(f"Object of type {type(obj)} is not JSON serializable")

        return json.dumps(asdict(data), default=convert_numpy)

    @staticmethod
    def generate_id(size: int = 6, chars: str = string.ascii_uppercase + string.digits) -> str:
        return "el" + "".join(random.choice(chars) for _ in range(size))

    def save_to_file(self, path_to_graph_dump_file: str) -> None:
        with open(path_to_graph_dump_file, "w") as graph_dump_file:
            graph_dump_file.write(self._to_json(self.serialize_state()))

    def save_layout(self, path_to_layout_dump_file: str) -> None:
        with open(path_to_layout_dump_file, "w") as layout_dump_file:
            layout_dump_file.write(self._to_json(self.nodes_points))

    @staticmethod
    def __build_serialized_state_from_unknown_data(data: Any) -> SerializedState:
        tracker = Tracker(**data.pop("tracker"))
        settings = SerializedSettings(**data.pop("settings"))
        state_changes = StateChanges(**data.pop("stateChanges"))
        recalculation_changes = StateChanges(**data.pop("recalculationChanges"))

        return SerializedState(
            nodes=data["nodes"],
            edges=data["edges"],
            tracker=tracker,
            settings=settings,
            stateChanges=state_changes,
            recalculationChanges=recalculation_changes,
        )

    def __load_serialized_state_from_graph_dump(self, path_to_graph_dump_file: str) -> SerializedState | None:
        # noinspection PyBroadException
        try:
            with open(path_to_graph_dump_file, "r") as graph_dump_file:
                s = self.__build_serialized_state_from_unknown_data(json.load(graph_dump_file))
                return s

        except Exception:
            warnings.warn(f"Failed to load graph dump")
            return None

    @staticmethod
    def __load_nodes_points_from_layout_dump(path_to_layout_dump_file: str) -> NodesPoints:
        result: NodesPoints = {}

        # noinspection PyBroadException
        try:
            with open(path_to_layout_dump_file, "r") as layout_dump_file:
                raw_points = json.load(layout_dump_file)

                for node_id, raw_point in raw_points.items():
                    x = raw_point.get("x")
                    y = raw_point.get("y")
                    if isinstance(x, (int, float)) and isinstance(y, (int, float)):
                        result[node_id] = NodePoint(x=x, y=y)
        except Exception:
            warnings.warn(f"Failed to load layout dump")

        return result

    def serialize_state(
        self,
    ) -> SerializedState:
        mapped_node_targets: dict[NodeId, NodeTargetId] = {
            node: target
            for target, nodes in self.targets.items()
            for node in (nodes if isinstance(nodes, list) else [nodes])
        }

        edgelist = self.edgelist.get_filtered_edges()
        nodelist = self.nodelist.nodelist_df.copy()

        source_col = edgelist.columns[0]
        target_col = edgelist.columns[1]

        # calc edge type
        # noinspection PyTypeChecker
        edgelist["type"] = edgelist.apply(
            lambda x: (
                mapped_node_targets.get(x[source_col])  # type: ignore
                if mapped_node_targets.get(x[source_col]) == "source"
                else mapped_node_targets.get(x[target_col]) or "suit"
            ),
            1,  # type: ignore
        )

        nodes_items = self._prepare_nodes(
            nodelist=nodelist,
            mapped_node_targets=mapped_node_targets,
            nodes_custom_colors=self.nodes_custom_colors,
        )

        edges_items = self._prepare_edges(
            edgelist=edgelist,
            nodes_items=nodes_items,
            edges_custom_colors=self.edges_custom_colors,
        )

        columns = [Column(id=col, name=col) for col in self.weight_cols]

        serialized_nodes_state = SerializedNodesState(
            normalizations=self._allowed_normalizations,
            selectedNormalizationId=self.nodes_norm_type if self.nodes_norm_type else "none",
            items=nodes_items,
            columns=columns,
            threshold=self.nodes_thresholds,
            weightsRange=self.nodes_weights_range,
            selectedWeightsColumnId=self.nodes_weight_col,
            targets=self._allowed_targets,
            sortField="name",  # TODO: Make it configurable
            sortOrder="asc",  # TODO: Make it configurable
        )

        serialized_edges_state = SerializedEdgesState(
            items=edges_items,
            normalizations=self._allowed_normalizations,
            selectedNormalizationId=self.edges_norm_type if self.edges_norm_type else "none",
            columns=columns,
            threshold=self.edges_thresholds,
            weightsRange=self.edges_weights_range,
            selectedWeightsColumnId=self.edges_weight_col,
        )

        # noinspection PyProtectedMember
        tracker = Tracker(
            hwid=RETE_CONFIG.user.pk,
            scope="transition_graph",
            eventstreamIndex=self.eventstream._eventstream_index,
        )

        settings = SerializedSettings(
            showEdgesWeightsOnCanvas=self.settings.show_weights,
            convertWeightsToPercents=self.settings.show_percents,
            showNodesNamesOnCanvas=self.settings.show_nodes_names,
            doNotFilterTargetNodes=self.settings.show_all_edges_for_targets,
            showNodesWithoutEdges=self.settings.show_nodes_without_links,
            showEdgesInfoOnHover=self.settings.show_edge_info_on_hover,
            openSidebarByDefault=self.settings.open_sidebar_by_default,
        )

        return SerializedState(
            tracker=tracker,
            restoreNodesPoints=all("x" in node and "y" in node for node in nodes_items),
            nodes=serialized_nodes_state,
            edges=serialized_edges_state,
            settings=settings,
            stateChanges=self._state_changes,
            recalculationChanges=self._recalculation_changes,
            # TODO: remove hardcode
            version="3.0.0",
        )

    @time_performance(
        scope="transition_graph",
        event_name="plot",
    )
    def plot(
        self,
        targets: TargetToNodesMap | None = None,
        edges_norm_type: NormType = None,
        nodes_threshold: ThresholdWithFallback | None = None,
        nodes_norm_type: NormType = None,
        edges_threshold: ThresholdWithFallback | None = None,
        nodes_weight_col: str | None = None,
        edges_weight_col: str | None = None,
        custom_weight_cols: list[str] | None = None,
        width: str | int | float | None = None,
        height: str | int | float | None = None,
        show_weights: bool | None = None,
        show_percents: bool | None = None,
        show_nodes_names: bool | None = None,
        show_all_edges_for_targets: bool | None = None,
        show_nodes_without_links: bool | None = None,
        show_edge_info_on_hover: bool | None = None,
        open_sidebar_by_default: bool | None = None,
        nodes_custom_colors: NodesCustomColors | None = None,
        edges_custom_colors: EdgesCustomColors | None = None,
        nodelist: Nodelist | pd.DataFrame | None = None,
        layout_dump: str | None = None,
        import_file: str | None = None,
    ) -> None:
        """
        Create interactive transition graph visualization with callback to sourcing eventstream.

        Parameters
        ----------
        edges_norm_type : {"full", "node", None}, default None
            Type of normalization that is used to calculate weights for graph edges.
            Based on ``edges_weight_col`` parameter the weight values are calculated.

            - If ``None``, normalization is not used, the absolute values are taken.
            - If ``full``, normalization across the whole eventstream.
            - If ``node``, normalization across each node (or outgoing transitions from each node).

            See :ref:`Transition graph user guide <transition_graph_weights>` for the details.

        nodes_norm_type : {"full", "node", None}, default None
            Currently not implemented. Always None.

        edges_weight_col : str, optional
            A column name from the :py:class:`.EventstreamSchema` which values will control the final
            edges' weights and displayed width as well.

            For each edge is calculated:

            - If ``None`` or ``user_id`` - the number of unique users.
            - If ``event_id`` - the number of transitions.
            - If ``session_id`` - the number of unique sessions.
            - If ``custom_col`` - the number of unique values in selected column.

            See :ref:`Transition graph user guide <transition_graph_weights>` for the details.

        edges_threshold : dict, optional
            Threshold mapping that defines the minimal weights for edges displayed on the canvas.

            - Keys should be of type str and contain the weight column names (the values from the
              :py:class:`.EventstreamSchema`).
            - Values of the dict are the thresholds for the edges that will be displayed.

            Support multiple weighting columns. In that case, logical OR will be applied.
            Edges with value less than at least one of thresholds will be hidden.
            Example: {'event_id': 100, user_id: 50}.

            See :ref:`Transition graph user guide<transition_graph_thresholds>` for the details.

        nodes_weight_col : str, optional
            A column name from the :py:class:`.EventstreamSchema` which values control the final
            nodes' weights and displayed diameter as well.

            For each node is calculated:

            - If ``None`` or ``user_id`` - the number of unique users.
            - If ``event_id`` - the number of events.
            - If ``session_id`` - the number of unique sessions.
            - If ``custom_col`` - the number of unique values in selected column.

            See :ref:`Transition graph user guide <transition_graph_weights>` for the details.

        nodes_threshold : dict, optional
            Threshold mapping that defines the minimal weights for nodes displayed on the canvas.

            - Keys should be of type str and contain the weight column names (the values from the
              :py:class:`.EventstreamSchema`).
            - Values of the dict are the thresholds for the nodes that will be displayed.
              They should be of type int or float.

            Support multiple weighting columns. In that case, logical OR will be applied.
            Nodes with value less than at least one of thresholds will be hidden.
            Example: {'event_id': 100, user_id: 50}.

            See :ref:`Transition graph user guide<transition_graph_thresholds>` for the details.

        custom_weight_cols : list of str, optional
            Custom columns from the :py:class:`.EventstreamSchema` that can be selected in ``edges_weight_col``
            and ``nodes_weight_col`` parameters. If ``session_col=session_id`` exists,
            it is added by default to this list.

        targets : dict, optional
            Events mapping that defines which nodes and edges should be colored for better visualization.

            - Possible keys: "positive" (green), "negative" (red), "source" (orange).
            - Possible values: list of events of a given type.

            See :ref:`Transition graph user guide<transition_graph_color_settings>` for the details.

        nodes_custom_colors : dict, optional
            Set nodes color explicitly. The dict keys are node names, the values are the corresponding colors.
            A color can be defined either as an HTML standard color name or a HEX code.
            See :ref:`Transition graph user guide<transition_graph_color_settings>` for the details.

        edges_custom_colors : dict, optional
            Set edges color explicitly. The dict keys are tuples of length 2, e.g. (path_start', 'catalog'),
            the values are the corresponding colors.
            A color can be defined either as an HTML standard color name or a HEX code.
            See :ref:`Transition graph user guide<transition_graph_color_settings>` for the details.

        width : str, int or float, default "100%"
            The width of the plot can be specified in the following ways:

            - In pixels (int or float);
            - In other CSS units (str). For example, the default value of "100%" means the plot will occupy 100%
              of the width of the Jupyter Notebook cell.

        height : str, int or float, default "60vh"
            The height of the plot can be specified as follows:

            - In pixels (int or float);
            - In other CSS units (str). For example, the default value of "60vh" means the plot will occupy 60%
              of the height of the browser window.

            The resulting height can't be lower than 600px.

        show_weights : bool, default True
            Hide/display the edge weight labels. By default, weights are shown.

        show_percents : bool, default False
            Display edge weights as percents. Available only if an edge normalization type is chosen.
            By default, weights are displayed in fractions.

        show_nodes_names : bool, default True
            Hide/display the node names. By default, names are shown.

        show_all_edges_for_targets : bool, default True
            This displaying option allows to ignore the threshold filters and always display
            any edge connected to a target node. By default, all such edges are shown.

        show_nodes_without_links : bool, default False
            Setting a threshold filter might remove all the edges connected to a node.
            Such isolated nodes might be considered as useless. This displaying option
            hides them in the canvas as well.

        show_edge_info_on_hover : bool, default True
            This parameter determines whether information about an edge (weight, source node, target node)
            is displayed when hovering the mouse over it.

        open_sidebar_by_default : bool, default True
            This parameter specifies whether the sidebar should be open by default when the interface is loaded.
            If set to True, the sidebar will be displayed upon initialization,
            providing users immediate access to its contents.
            If set to False, the sidebar will remain hidden until the user decides to open it.

        nodelist : pd.DataFrame, default None
            A DataFrame containing information about nodes, such as weights, parents, etc.

        layout_dump : str, default None
            A string path to a JSON file containing the configuration for nodes layout.

        import_file : str, default None
            A string path to a JSON file containing complete dump of a graph.

        Returns
        -------
            Rendered IFrame graph.

        Notes
        -----
        1. If all the edges connected to a node are hidden, the node becomes hidden as well.
           In order to avoid it - use ``show_nodes_without_links=True`` parameter in code or in the interface.
        2. The thresholds may use their own weighting columns both for nodes and for edges independently
           of weighting columns defined in ``edges_weight_col`` and ``nodes_weight_col`` arguments.

        See :doc:`TransitionGraph user guide </user_guides/transition_graph>` for the details.
        """
        if edges_norm_type is None and show_percents:
            raise ValueError("If show_percents=True, edges_norm_type should be 'full' or 'node'!")

        self.__prepare_state_for_plot(
            targets=targets,
            edges_norm_type=edges_norm_type,
            nodes_threshold=nodes_threshold,
            nodes_norm_type=nodes_norm_type,
            edges_threshold=edges_threshold,
            nodes_weight_col=nodes_weight_col,
            edges_weight_col=edges_weight_col,
            custom_weight_cols=custom_weight_cols,
            width=width,
            height=height,
            show_weights=show_weights,
            show_percents=show_percents,
            show_nodes_names=show_nodes_names,
            show_all_edges_for_targets=show_all_edges_for_targets,
            show_nodes_without_links=show_nodes_without_links,
            show_edge_info_on_hover=show_edge_info_on_hover,
            open_sidebar_by_default=open_sidebar_by_default,
            nodes_custom_colors=nodes_custom_colors,
            edges_custom_colors=edges_custom_colors,
            nodelist=nodelist,
            path_to_layout_dump_file=layout_dump,
            path_to_graph_dump_file=import_file,
        )

        widget_id = self.generate_id()

        serialized_state = self.serialize_state()

        display(
            HTML(
                self.render.show(
                    script_url=f"{self.graph_url}?id={widget_id}",
                    style=f"width: 100%; width: {self.width}; height: 60vh; height: {self.height}; min-height: 600px; box-sizing: border-box;",
                    state=self._to_json(serialized_state),
                    server=self.server,
                    bridge_model_id=self.bridge_id,
                    id=f"graph-{widget_id}",
                )
            )
        )

        # noinspection PyProtectedMember
        collect_data_performance(
            scope="transition_graph",
            event_name="metadata",
            called_params={
                "edges_norm_type": edges_norm_type,
                "nodes_norm_type": nodes_norm_type,
                "targets": targets,
                "nodes_threshold": nodes_threshold,
                "edges_threshold": edges_threshold,
                "nodes_weight_col": nodes_weight_col,
                "edges_weight_col": edges_weight_col,
                "custom_weight_cols": custom_weight_cols,
                "width": width,
                "height": height,
                "show_weights": show_weights,
                "show_percents": show_percents,
                "show_nodes_names": show_nodes_names,
                "show_all_edges_for_targets": show_all_edges_for_targets,
                "show_nodes_without_links": show_nodes_without_links,
                "show_edge_info_on_hover": show_edge_info_on_hover,
                "open_sidebar_by_default": open_sidebar_by_default,
            },
            not_hash_values=["edges_norm_type", "targets", "width", "height"],
            performance_data={
                "unique_nodes": len(serialized_state.nodes["items"]),
                "unique_links": len(serialized_state.edges["items"]),
            },
            eventstream_index=self.eventstream._eventstream_index,
        )

    def __prepare_state_for_plot(
        self,
        targets: TargetToNodesMap | None = None,
        edges_norm_type: NormType = None,
        nodes_threshold: ThresholdWithFallback | None = None,
        nodes_norm_type: NormType = None,
        edges_threshold: ThresholdWithFallback | None = None,
        nodes_weight_col: str | None = None,
        edges_weight_col: str | None = None,
        custom_weight_cols: list[str] | None = None,
        width: str | int | float | None = None,
        height: str | int | float | None = None,
        show_weights: bool | None = None,
        show_percents: bool | None = None,
        show_nodes_names: bool | None = None,
        show_all_edges_for_targets: bool | None = None,
        show_nodes_without_links: bool | None = None,
        show_edge_info_on_hover: bool | None = None,
        open_sidebar_by_default: bool | None = None,
        nodes_custom_colors: NodesCustomColors | None = None,
        edges_custom_colors: EdgesCustomColors | None = None,
        nodelist: Nodelist | pd.DataFrame | None = None,
        path_to_layout_dump_file: str | None = None,
        path_to_graph_dump_file: str | None = None,
    ) -> None:
        serialized_state_from_file = (
            self.__load_serialized_state_from_graph_dump(path_to_graph_dump_file) if path_to_graph_dump_file else None
        )

        if path_to_layout_dump_file:
            self.nodes_points = self.__load_nodes_points_from_layout_dump(path_to_layout_dump_file)
        elif serialized_state_from_file:
            self.nodes_points = self._extract_nodes_points_from_serialized_state(serialized_state_from_file)

        if serialized_state_from_file:
            self._state_changes = serialized_state_from_file.stateChanges
            self._recalculation_changes = serialized_state_from_file.recalculationChanges

        if show_weights is not None:
            self.settings.show_weights = show_weights
        elif serialized_state_from_file:
            self.settings.show_weights = serialized_state_from_file.settings.showEdgesWeightsOnCanvas

        if show_percents is not None:
            self.settings.show_percents = show_percents
        elif serialized_state_from_file:
            self.settings.show_percents = serialized_state_from_file.settings.convertWeightsToPercents

        if show_nodes_names is not None:
            self.settings.show_nodes_names = show_nodes_names
        elif serialized_state_from_file:
            self.settings.show_nodes_names = serialized_state_from_file.settings.showNodesNamesOnCanvas

        if show_all_edges_for_targets is not None:
            self.settings.show_all_edges_for_targets = show_all_edges_for_targets
        elif serialized_state_from_file:
            self.settings.show_all_edges_for_targets = serialized_state_from_file.settings.doNotFilterTargetNodes

        if show_nodes_without_links is not None:
            self.settings.show_nodes_without_links = show_nodes_without_links
        elif serialized_state_from_file:
            self.settings.show_nodes_without_links = serialized_state_from_file.settings.showNodesWithoutEdges

        if show_edge_info_on_hover is not None:
            self.settings.show_edge_info_on_hover = show_edge_info_on_hover
        elif serialized_state_from_file:
            self.settings.show_edge_info_on_hover = serialized_state_from_file.settings.showEdgesInfoOnHover

        if open_sidebar_by_default is not None:
            self.settings.open_sidebar_by_default = open_sidebar_by_default
        elif serialized_state_from_file:
            self.settings.open_sidebar_by_default = serialized_state_from_file.settings.openSidebarByDefault

        self.targets = {}
        if targets:
            self.targets = targets
        elif serialized_state_from_file:
            targets_result: TargetToNodesMap = {}
            for node in serialized_state_from_file.nodes["items"]:
                target_id = node.get("targetId")
                if target_id:
                    if target_id not in targets_result:
                        targets_result[target_id] = []
                    targets_result[target_id].append(node["id"])  # type: ignore
            self.targets = targets_result

        self.nodes_norm_type = None
        if nodes_norm_type:
            self.nodes_norm_type = nodes_norm_type
        elif serialized_state_from_file:
            serialized_nodes_norm_type = serialized_state_from_file.nodes["selectedNormalizationId"]
            if serialized_nodes_norm_type != "none":
                self.nodes_norm_type = serialized_nodes_norm_type

        self.edges_norm_type = None
        if edges_norm_type:
            self.edges_norm_type = edges_norm_type
        elif serialized_state_from_file:
            serialized_edges_norm_type = serialized_state_from_file.edges["selectedNormalizationId"]
            if serialized_edges_norm_type != "none":
                self.edges_norm_type = serialized_edges_norm_type

        self.nodelist_default_col = self.eventstream.schema.event_id
        self.edgelist_default_col = self.eventstream.schema.event_id

        self.weight_cols = self._define_weight_cols(custom_weight_cols)

        if nodes_weight_col:
            self.nodes_weight_col = nodes_weight_col
        elif serialized_state_from_file:
            self.nodes_weight_col = serialized_state_from_file.nodes["selectedWeightsColumnId"]
        else:
            self.nodes_weight_col = self.eventstream.schema.user_id

        if edges_weight_col:
            self.edges_weight_col = edges_weight_col
        elif serialized_state_from_file:
            self.edges_weight_col = serialized_state_from_file.edges["selectedWeightsColumnId"]
        else:
            self.edges_weight_col = self.eventstream.schema.user_id

        if width is not None:
            self.width = f"{width}px" if isinstance(width, (int, float)) else width

        if height is not None:
            self.height = f"{height}px" if isinstance(height, (int, float)) else height

        self.nodes_custom_colors = {}
        if nodes_custom_colors:
            self.nodes_custom_colors = nodes_custom_colors
        elif serialized_state_from_file:
            nodes_custom_colors_result: NodesCustomColors = {}
            for node in serialized_state_from_file.nodes["items"]:
                custom_color = node.get("customColor")
                if custom_color:
                    nodes_custom_colors_result[node["id"]] = custom_color
            self.nodes_custom_colors = nodes_custom_colors_result

        self.edges_custom_colors = {}
        if edges_custom_colors:
            self.edges_custom_colors = edges_custom_colors
        elif serialized_state_from_file:
            edges_custom_colors_result: EdgesCustomColors = {}
            for edge in serialized_state_from_file.edges["items"]:
                custom_color = edge.get("customColor")
                if custom_color:
                    edges_custom_colors_result[(edge["sourceNodeId"], edge["targetNodeId"])] = custom_color
            self.edges_custom_colors = edges_custom_colors_result

        # calculate nodelist & edgelist on initial eventstream without any thresholds
        self.initial_nodelist: Nodelist = Nodelist(
            weight_cols=self.weight_cols,
            time_col=self.event_time_col,
            event_col=self.event_col,
        )
        self.initial_nodelist.calculate_nodelist(data=self.eventstream.to_dataframe())

        self.initial_edgelist: Edgelist = Edgelist(eventstream=self.eventstream)
        self.initial_edgelist.calculate_edgelist(
            weight_cols=self.weight_cols,
            norm_type=self.edges_norm_type,
        )

        self.nodes_weights_range = self.initial_nodelist.get_min_max()
        self.edges_weights_range = self.initial_edgelist.get_min_max()

        self.nodes_thresholds = {}
        if nodes_threshold:
            self.nodes_thresholds = self.__threshold_fallback(nodes_threshold, threshold_type="nodes")
        elif serialized_state_from_file:
            self.nodes_thresholds = serialized_state_from_file.nodes.get("threshold", {})

        self.edges_thresholds = {}
        if edges_threshold:
            self.edges_thresholds = self.__threshold_fallback(edges_threshold, threshold_type="edges")
        elif serialized_state_from_file:
            self.edges_thresholds = serialized_state_from_file.edges.get("threshold", {})

        external_nodelist_df: pd.DataFrame | None = None
        if nodelist:
            external_nodelist_df = nodelist if isinstance(nodelist, pd.DataFrame) else nodelist.nodelist_df
        elif serialized_state_from_file:
            external_nodelist_df = pd.DataFrame(self.make_nodelist_from_serialized_state(serialized_state_from_file))
        self.__apply_nodelist(nodelist_df=external_nodelist_df)

    def __apply_nodelist(self, nodelist_df: pd.DataFrame | None = None) -> EventstreamType:
        prev_nodes_min_max = (
            self.nodelist.get_min_max()
            if hasattr(self, "nodelist") and self.nodelist is not None  # type: ignore
            else None
        )

        prev_edges_min_max = (
            self.edgelist.get_min_max()
            if hasattr(self, "edgelist") and self.edgelist is not None  # type: ignore
            else None
        )

        self.nodelist: Nodelist = self.initial_nodelist.copy()
        if nodelist_df is not None:
            self.nodelist.update(nodelist_df)

        curr_stream = self.eventstream

        disabled_nodes = self.nodelist.get_disabled_nodes()
        if len(disabled_nodes) > 0:
            curr_stream = curr_stream.filter_events(
                func=lambda df, schema: ~df[schema.event_name].isin(disabled_nodes)  # type: ignore
            )

        rename_rules = self.nodelist.groups_to_rename_rules()
        if len(rename_rules) > 0:
            curr_stream = curr_stream.rename(rules=rename_rules)  # type: ignore

        self.nodelist.calculate_nodelist(data=curr_stream.to_dataframe())

        if prev_nodes_min_max is not None:
            self.nodes_thresholds = self.nodelist.fit_threshold(
                threshold=self.nodes_thresholds, prev_min_max=prev_nodes_min_max
            )

        self.nodelist.update_threshold(nodes_thresholds=self.nodes_thresholds)

        out_of_threshold_nodes = self.nodelist.get_out_of_threshold_nodes(only_ungrouped=True)
        if len(out_of_threshold_nodes) > 0:
            curr_stream = curr_stream.filter_events(  # type: ignore
                func=lambda df, schema: ~df[schema.event_name].isin(out_of_threshold_nodes)
            )

        self.edgelist: Edgelist = Edgelist(eventstream=curr_stream)
        self.edgelist.calculate_edgelist(
            weight_cols=self.weight_cols,
            norm_type=self.edges_norm_type,
        )

        if prev_edges_min_max is not None:
            self.edges_thresholds = self.edgelist.fit_threshold(
                threshold=self.edges_thresholds, prev_min_max=prev_edges_min_max
            )

        self.edgelist.update_threshold(edges_threshold=self.edges_thresholds)

        # use aliases after all calculations
        aliases_rename_rules = self.nodelist.renamed_events_to_rename_rules()
        if len(aliases_rename_rules) > 0:
            curr_stream = curr_stream.rename(rules=aliases_rename_rules)  # type: ignore

        return curr_stream

    def __threshold_fallback(
        self, threshold: ThresholdWithFallback, threshold_type: Literal["nodes"] | Literal["edges"]
    ) -> ThresholdValueMap:
        result: ThresholdValueMap = {}

        for key, value in threshold.items():
            weights_range = (
                self.nodes_weights_range.get(key, None)
                if threshold_type == "nodes"
                else self.edges_weights_range.get(key, None)
            )

            if weights_range is None:
                raise ValueError(f"threshold is invalid. Column '{key}' doesn't found!")

            if isinstance(value, dict):
                result[key] = value
                continue

            result[key] = {"min": float(value), "max": float(weights_range["max"])}

        return result



from __future__ import annotations

import json
import os
import uuid
import warnings
from dataclasses import asdict, field
from pathlib import Path
from typing import Optional, Type, Union

from pydantic import BaseConfig
from pydantic.dataclasses import dataclass

from retentioneering.utils.hwid import get_hwid  # type: ignore


@dataclass
class TrackingConfig(BaseConfig):
    """
    Configuration for tracking.
    """

    is_tracking_allowed: bool = True


@dataclass
class UserConfig(BaseConfig):
    pk: str = ""

    def __post_init__(self) -> None:
        if self.pk == "":
            self.pk = str(uuid.uuid4())


@dataclass
class TransitionGraphConfig(BaseConfig):
    """
    Configuration for the transition graph.
    """

    width: Union[float, int, str] = "100%"
    height: Union[float, int, str] = "800px"
    show_weights: bool = True
    show_percents: bool = False
    show_nodes_names: bool = True
    show_all_edges_for_targets: bool = True
    show_nodes_without_links: bool = False
    show_edge_info_on_hover: bool = True
    open_sidebar_by_default: bool = True


@dataclass
class PreprocessingGraphConfig(BaseConfig):
    """
    Configuration for the preprocessing graph.
    """

    width: int = 960
    height: int = 600


DEFAULT_CONFIG = {
    "user": asdict(UserConfig()),
    "transition_graph": asdict(TransitionGraphConfig()),
    "preprocessing_graph": asdict(PreprocessingGraphConfig()),
}


@dataclass
class Config(BaseConfig):
    """
    Configuration for the application.
    """

    tracking: TrackingConfig = field(default_factory=TrackingConfig)
    user: UserConfig = field(default_factory=UserConfig)
    transition_graph: TransitionGraphConfig = field(default_factory=TransitionGraphConfig)
    preprocessing_graph: PreprocessingGraphConfig = field(default_factory=PreprocessingGraphConfig)

    def _inner_mapping(self, _validate: bool = True) -> dict[str, Type[BaseConfig]]:
        # !!! DO NOT SET _validate to False except for tests !!!
        # self.__annotations__ has type dict[str, str], and it will be simpler than obtain class by name from globals or something like that
        # And it's method because of pydantic dataclasses
        mapping = {
            "tracking": TrackingConfig,
            "user": UserConfig,
            "transition_graph": TransitionGraphConfig,
            "preprocessing_graph": PreprocessingGraphConfig,
        }

        if _validate and (missing_keys := set(self.__annotations__.keys()) - set(mapping.keys())):
            raise ValueError(f"The keys of __annotations__ and inner mapping are not the same. Lost {missing_keys}")
        return mapping  # type: ignore

    @staticmethod
    def _get_path_for_config() -> Optional[str]:
        config_filename = ".retentioneering_config.json"
        home = str(Path.home())
        if os.access(f"{home}", os.W_OK):
            return f"{home}/{config_filename}"
        elif os.access(".", os.W_OK):
            return config_filename
        else:
            return None

    def __post_init__(self) -> None:
        config_path = self._get_path_for_config()
        if config_path is None:
            return None
        # check if file exists
        if not os.path.isfile(config_path):
            with open(config_path, "w") as f:
                json.dump(DEFAULT_CONFIG, f)
        self.load()

    def load(self) -> None:
        config_path = self._get_path_for_config()
        if config_path is None:
            return None
        with open(config_path, "r") as f:
            try:
                config: dict[str, dict] = json.load(f)
                mapping = self._inner_mapping()
                for section, _class in mapping.items():
                    section_data = config.get(section, {})
                    setattr(self, section, _class(**section_data))  # type: ignore

            except json.decoder.JSONDecodeError:
                warnings.warn("Invalid config file. Please fix it and try again. Current config file is default.")

            else:
                # update config file from actual state
                current_config_data = asdict(self)
                if config.get("tracking", None) is None:
                    del current_config_data["tracking"]

                with open(config_path, "w") as f:
                    json.dump(current_config_data, f)

    def save(self) -> None:
        config_path = self._get_path_for_config()
        if config_path is None:
            return None

        self._write_current_config(config_path)

    def _write_current_config(self, config_path: str) -> None:
        sections = self._inner_mapping().keys()
        config_data = {section: asdict(getattr(self, section)) for section in sections}
        with open(config_path, "w") as f:
            json.dump(config_data, f, indent=4)


RETE_CONFIG = Config()



from __future__ import annotations

from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.common.constants import DATETIME_UNITS
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ListOfIds, ReteTimeWidget


class LabelLostUsersParams(ParamsModel):
    """
    A class with parameters for :py:class:`.LabelLostUsers` class.
    """

    timeout: Optional[Tuple[float, DATETIME_UNITS]]
    lost_users_list: Optional[Union[List[int], List[str]]]

    _widgets = {
        "timeout": ReteTimeWidget(),
        "lost_users_list": ListOfIds(),
    }


@docstrings.get_sections(base="LabelLostUsers")  # type: ignore
class LabelLostUsers(DataProcessor):
    """
    Create one of synthetic events in each user's path:
    ``lost_user`` or ``absent_user``.


    Parameters
    ----------
    Only one of parameters could be used at the same time
    timeout : Tuple(float, :numpy_link:`DATETIME_UNITS<>`), optional
        Threshold value and its unit of measure.
        Calculate timedelta between the last event in each user's path and the last event in the whole eventstream.
        For users with timedelta greater or equal to selected ``timeout``, a new synthetic event - ``lost_user``
        will be added.
        For other users paths a new synthetic event - ``absent_user`` will be added.
    lost_users_list : list of int or list of str, optional
        If the `list of user_ids` is given new synthetic event - ``lost_user`` will be added to each user from the list.
        For other user's paths will be added new synthetic event - ``absent_user``.

    Returns
    -------
    Eventstream
        ``Eventstream`` with new synthetic events only - one for each user:

        +-----------------+-----------------+------------------+
        | **event_name**  | **event_type**  |  **timestamp**   |
        +-----------------+-----------------+------------------+
        | lost_user       | lost_user       | last_event       |
        +-----------------+-----------------+------------------+
        | absent_user     | absent_user     | last_event       |
        +-----------------+-----------------+------------------+

    Raises
    ------
    ValueError
        Raised when both ``timeout`` and ``lost_users_list`` are either empty or given.

    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.
    """

    params: LabelLostUsersParams

    @time_performance(
        scope="label_lost_users",
        event_name="init",
    )
    def __init__(self, params: LabelLostUsersParams):
        super().__init__(params=params)

    @time_performance(  # type: ignore
        scope="label_lost_users",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        user_col = schema.user_id
        time_col = schema.event_timestamp
        type_col = schema.event_type
        event_col = schema.event_name

        timeout, timeout_unit = None, None
        lost_users_list = self.params.lost_users_list
        data_lost = pd.DataFrame()

        if self.params.timeout:
            timeout, timeout_unit = self.params.timeout

        if timeout and lost_users_list:
            raise ValueError("timeout and lost_users_list parameters cannot be used simultaneously!")

        if not timeout and not lost_users_list:
            raise ValueError("Either timeout or lost_users_list must be specified!")

        if timeout and timeout_unit:
            data_lost = df.groupby(user_col, as_index=False).last()
            data_lost["diff_end_to_end"] = data_lost[time_col].max() - data_lost[time_col]

            data_lost["diff_end_to_end"] /= np.timedelta64(1, timeout_unit)  # type: ignore

            data_lost[type_col] = np.where(data_lost["diff_end_to_end"] < timeout, "absent_user", "lost_user")
            data_lost[event_col] = data_lost[type_col]
            del data_lost["diff_end_to_end"]

        if lost_users_list:
            data_lost = df.groupby(user_col, as_index=False).last()
            data_lost[type_col] = np.where(data_lost["user_id"].isin(lost_users_list), "lost_user", "absent_user")
            data_lost[event_col] = data_lost[type_col]

        result = pd.concat([df, data_lost])

        collect_data_performance(
            scope="label_lost_users",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            not_hash_values=["timeout"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



from __future__ import annotations

from typing import Callable, List

import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType, EventstreamType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ListOfString, ReteFunction


def _default_func(eventstream: EventstreamType, targets: list[str]) -> pd.DataFrame:
    """
    Filter rows with target events from the input eventstream.

    Parameters
    ----------
    eventstream : Eventstream
        Source eventstream or output from previous nodes.
    targets : list of str
        Condition for eventstream filtering.
        Each event from that list is associated with a conversion goal of the user behaviour in the product.
        If there are several target events in user path - the event with minimum timestamp is taken.

    Returns
    -------
    pd.DataFrame
        Filtered DataFrame with targets and its timestamps.
    """
    user_col = eventstream.schema.user_id
    time_col = eventstream.schema.event_timestamp
    event_col = eventstream.schema.event_name
    df = eventstream.to_dataframe()

    targets_index = df[df[event_col].isin(targets)].groupby(user_col)[time_col].idxmin()  # type: ignore

    return df.loc[targets_index]  # type: ignore


class AddPositiveEventsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.AddPositiveEvents` class.
    """

    targets: List[str]
    func: Callable = _default_func

    _widgets = {"func": ReteFunction(), "targets": ListOfString()}


@docstrings.get_sections(base="AddPositiveEvents")  # type: ignore
class AddPositiveEvents(DataProcessor):
    """
    Create new synthetic events in paths of all users having the specified events:
    ``positive_target_RAW_EVENT_NAME``

    Parameters
    ----------
    targets : list of str
        Define the list of events we consider positive.
        If there are several target events in a user path, the event with the minimum timestamp is taken.
    func : Callable, default _default_func
        Filter rows with target events from the input eventstream.

    Returns
    -------
    Eventstream
        ``Eventstream`` with new synthetic events only added to users who fit the conditions.

        +--------------------------------+-----------------+-----------------------------+
        | **event_name**                 | **event_type**  | **timestamp**               |
        +--------------------------------+-----------------+-----------------------------+
        | positive_target_RAW_EVENT_NAME | positive_target | min(targets)                |
        +--------------------------------+-----------------+-----------------------------+

    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.
    """

    params: AddPositiveEventsParams

    @time_performance(
        scope="add_positive_events",
        event_name="init",
    )
    def __init__(self, params: AddPositiveEventsParams):
        super().__init__(params=params)

    @time_performance(
        scope="add_positive_events",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        from retentioneering.eventstream.eventstream import Eventstream

        type_col = schema.event_type
        event_col = schema.event_name

        func: Callable[[EventstreamType, list[str]], pd.DataFrame] = self.params.func
        targets = self.params.targets

        eventstream = Eventstream(
            raw_data_schema=schema.to_raw_data_schema(event_index=True),
            raw_data=df,
            add_start_end_events=False,
        )

        positive_targets = func(eventstream, targets)
        positive_targets[type_col] = "positive_target"
        positive_targets[event_col] = "positive_target_" + positive_targets[event_col]

        result = pd.concat([eventstream.to_dataframe(), positive_targets])

        collect_data_performance(
            scope="add_positive_events",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



from __future__ import annotations

from dataclasses import dataclass, field

from .edges import RecalculationEdgeItem
from .nodes import NodeId, RecalculationNodeItem
from .threshold import ThresholdValueMap


@dataclass
class RecalculationResult:
    nodes: dict[NodeId, RecalculationNodeItem] = field(default_factory=dict)
    edges: list[RecalculationEdgeItem] = field(default_factory=list)
    nodesThresholds: ThresholdValueMap = field(default_factory=dict)
    edgesThresholds: ThresholdValueMap = field(default_factory=dict)
    nodesThresholdsMinMax: ThresholdValueMap = field(default_factory=dict)
    edgesThresholdsMinMax: ThresholdValueMap = field(default_factory=dict)



# * Copyright (C) 2020 Maxim Godzi, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

# TODO fix me
from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Optional

from retentioneering.eventstream.types import (
    EventstreamSchemaType,
    RawDataCustomColSchema,
    RawDataSchemaType,
)


@dataclass
class EventstreamSchema(EventstreamSchemaType):
    """
    Define a schema for ``eventstream`` columns names.
    If names of the columns are different from default names, they need to be
    specified.

    Parameters
    ----------
    event_id : str, default "event_id"
    event_type : str, default "event_type"
    event_index : str, default "event_index"
    event_name : str, default "event"
    event_timestamp : str, default "timestamp"
    user_id : str, default "user_id"
    custom_cols : list of str, optional

    Notes
    -----
    See :ref:`Eventstream user guide<eventstream_field_names>` for the details.

    """

    event_id: str = "event_id"
    event_type: str = "event_type"
    event_index: str = "event_index"
    event_name: str = "event"
    event_timestamp: str = "timestamp"
    user_id: str = "user_id"
    custom_cols: List[str] = field(default_factory=list)

    def copy(self) -> EventstreamSchema:
        return EventstreamSchema(
            event_id=self.event_id,
            event_type=self.event_type,
            event_index=self.event_index,
            event_name=self.event_name,
            event_timestamp=self.event_timestamp,
            user_id=self.user_id,
            custom_cols=self.custom_cols.copy(),
        )

    def is_equal(self, schema: EventstreamSchemaType) -> bool:
        return (
            self.event_id == schema.event_id
            and self.event_type == schema.event_type
            and self.event_index == schema.event_index
            and self.event_name == schema.event_name
            and self.event_timestamp == schema.event_timestamp
            and self.user_id == schema.user_id
            and (set(self.custom_cols).issubset(schema.custom_cols))
        )

    def get_default_cols(self) -> List[str]:
        return [
            self.event_id,
            self.event_type,
            self.event_index,
            self.event_name,
            self.event_timestamp,
            self.user_id,
        ]

    def get_cols(self) -> list[str]:
        return [
            self.event_id,
            self.event_type,
            self.event_index,
            self.event_name,
            self.event_timestamp,
            self.user_id,
        ] + self.custom_cols

    def to_raw_data_schema(self, event_id: bool = False, event_index: bool = False) -> RawDataSchema:
        custom_cols: List[RawDataCustomColSchema] = []

        for col in self.custom_cols:
            custom_cols.append({"custom_col": col, "raw_data_col": col})

        return RawDataSchema(
            event_name=self.event_name,
            event_type=self.event_type,
            event_id=self.event_id if event_id else None,
            event_index=self.event_index if event_index else None,
            user_id=self.user_id,
            event_timestamp=self.event_timestamp,
            custom_cols=custom_cols,
        )


@dataclass
class RawDataSchema(RawDataSchemaType):
    """
    Define schema for ``raw_data`` columns names.
    If names of the columns are different from default names, they need to be
    specified.

    Parameters
    ----------
    event_name : str, default "event"
    event_timestamp : str, default "timestamp"
    user_id : str, default "user_id"
    event_type : str, optional
    event_index: str, optional
    custom_cols : list, optional

    Notes
    -----
    See :ref:`Eventstream user guide<eventstream_raw_data_schema>` for the details.

    """

    event_name: str = "event"
    event_timestamp: str = "timestamp"
    user_id: str = "user_id"
    event_index: Optional[str] = None
    event_type: Optional[str] = None
    event_id: Optional[str] = None
    custom_cols: List[RawDataCustomColSchema] = field(default_factory=list)

    def get_default_cols(self) -> List[str]:
        cols: List[str] = [
            self.event_name,
            self.event_timestamp,
            self.user_id,
        ]

        if self.event_index:
            cols.append(self.event_index)

        if self.event_type:
            cols.append(self.event_type)

        if self.event_id:
            cols.append(self.event_id)

        return cols

    def copy(self) -> RawDataSchema:
        return RawDataSchema(
            event_name=self.event_name,
            event_timestamp=self.event_timestamp,
            user_id=self.user_id,
            custom_cols=self.custom_cols,
            event_type=self.event_type,
            event_index=self.event_index,
            event_id=self.event_id,
        )



from .add_negative_events_helper import AddNegativeEventsHelperMixin
from .add_positive_events_helper import AddPositiveEventsHelperMixin
from .add_segment_helper import AddSegmentHelperMixin
from .add_start_end_events_helper import AddStartEndEventsHelperMixin
from .collapse_loops_helper import CollapseLoopsHelperMixin
from .drop_paths_helper import DropPathsHelperMixin
from .drop_segment_helper import DropSegmentHelperMixin
from .filter_events_helper import FilterEventsHelperMixin
from .group_events_bulk_helper import GroupEventsBulkHelperMixin
from .group_events_helper import GroupEventsHelperMixin
from .label_cropped_paths_helper import LabelCroppedPathsHelperMixin
from .label_lost_users_helper import LabelLostUsersHelperMixin
from .label_new_users_helper import LabelNewUsersHelperMixin
from .materialize_segment_helper import MaterializeSegmentHelperMixin
from .pipe_helper import PipeHelperMixin
from .remap_segment_helper import RemapSegmentHelperMixin
from .rename_helper import RenameHelperMixin
from .rename_segment_helper import RenameSegmentHelperMixin
from .split_sessions_helper import SplitSessionsHelperMixin
from .truncate_paths_helper import TruncatePathsHelperMixin



# TODO fix me
from __future__ import annotations

from typing import Callable, List, Optional, TypeVar

T = TypeVar("T")


def find_item(input_list: List[T], cond: Callable[[T], bool]) -> Optional[T]:
    for item in input_list:
        if cond(item):
            return item
    return None


def find_index(input_list: List[T], cond: Callable[[T], bool]) -> int:
    for i, item in enumerate(input_list):
        if cond(item):
            return i
    return -1



from __future__ import annotations

import inspect
import types
from dataclasses import dataclass, field
from typing import Any, Callable, List, Type, Union

from retentioneering.common.constants import DATETIME_UNITS_LIST
from retentioneering.exceptions.widget import ParseReteFuncError


@dataclass
class StringWidget:
    default: str | None = None
    widget: str = "string"

    @classmethod
    def from_dict(cls: Type[StringWidget], **kwargs: Any) -> "StringWidget":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})


@dataclass
class IntegerWidget:
    default: int | None = None
    widget: str = "integer"

    @classmethod
    def from_dict(cls: Type[IntegerWidget], **kwargs: Any) -> "IntegerWidget":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})


@dataclass
class EnumWidget:
    default: Any | None = None
    params: list[str] | None = None
    widget: str = "enum"
    optional: bool | None = None

    @classmethod
    def from_dict(cls: Type[EnumWidget], **kwargs: Any) -> "EnumWidget":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def _serialize(cls, value: str) -> str:
        return value

    @classmethod
    def _parse(cls, value: str) -> str:  # type: ignore
        return value


@dataclass
class BooleanWidget:
    default: bool = False
    widget: str = "boolean"

    @classmethod
    def from_dict(cls: Type[BooleanWidget], **kwargs: Any) -> "BooleanWidget":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})


@dataclass
class ReteTimeWidget:
    default: tuple[float, str] | None = None
    widget: str = "time_widget"
    params: list = field(
        default_factory=lambda: [{"widget": "float"}, {"params": DATETIME_UNITS_LIST, "widget": "enum"}]
    )

    @classmethod
    def from_dict(cls: Type[ReteTimeWidget], **kwargs: Any) -> "ReteTimeWidget":
        kwargs["params"] = [
            {"widget": "float"},
            {"widget": "enum", "params": ["Y", "M", "W", "D", "h", "m", "s", "ms", "us", "μs", "ns", "ps", "fs", "as"]},
        ]
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def _serialize(cls, value: tuple[float, str]) -> tuple[float, str]:
        return value

    @classmethod
    def _parse(cls: Type[ReteTimeWidget], value: str) -> tuple[float, str]:  # type: ignore
        TIME, QUANT = 0, 1

        if type(value) is tuple:
            return value  # type: ignore

        if type(value) is list and len(value) != 2:
            raise Exception("Incorrect input")

        if type(value) is list:
            return float(value[TIME]), str(value[QUANT])

        data = value.split(",")

        if len(data) != 2:
            raise Exception("Incorrect input")
        return float(data[TIME]), str(data[QUANT])


@dataclass
class ReteFunction:
    default: Callable | None = None
    widget: str = "function"

    @classmethod
    def from_dict(cls: Type[ReteFunction], **kwargs: Any) -> "ReteFunction":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def _serialize(cls, value: Callable) -> str:
        try:
            code = inspect.getsource(value)
            return code
        except OSError:
            return getattr(value, "_source_code", "")

    @classmethod
    def _parse(cls, value: str) -> Callable:  # type: ignore
        func_str = value.strip()
        try:
            code_obj = compile(func_str, "<string>", "exec")
        except:
            raise ParseReteFuncError("parsing error. You must implement a python function here")

        new_func_type = None

        for i in code_obj.co_consts:
            try:
                new_func_type = types.FunctionType(i, globals=globals())
                break
            except TypeError as err:
                continue

        if new_func_type is None:
            raise ParseReteFuncError("parsing error. You must implement a python function here")

        setattr(new_func_type, "_source_code", func_str)
        return new_func_type


@dataclass
class ListOfInt:
    default: list[int] | None = None
    widget: str = "list_of_int"

    @classmethod
    def from_dict(cls: Type[ListOfInt], **kwargs: Any) -> "ListOfInt":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def _serialize(cls: Type[ListOfInt], value: list[int] | None) -> list[int] | None:
        return value

    @classmethod
    def _parse(cls: Type[ListOfInt], value: list[int]) -> list[int] | None:  # type: ignore
        return value


@dataclass
class ListOfUsers:
    # @TODO: remove this widget and make his functionality in ListOfInt. Vladimir Makhanov
    default: list[int] | list[str] | None = None
    params: dict[str, str] | None = field(default_factory=lambda: {"disable_value": "all"})
    widget: str = "list_of_ids"

    @classmethod
    def from_dict(cls: Type[ListOfUsers], **kwargs: Any) -> "ListOfUsers":
        kwargs["params"] = {"disable_value": "all"}
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def _serialize(cls: Type[ListOfUsers], value: list[int] | list[str] | None) -> list[int] | list[str] | None:
        return value

    @classmethod
    def _parse(cls: Type[ListOfUsers], value: list[int] | list[str]) -> list[int] | list[str] | None:  # type: ignore
        return value


@dataclass
class ListOfIds:
    default: list[int] | list[str] | None = None
    widget: str = "list_of_ids"

    @classmethod
    def from_dict(cls: Type[ListOfIds], **kwargs: Any) -> "ListOfIds":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def _serialize(cls: Type[ListOfIds], value: list[int] | list[str] | None) -> list[int] | list[str] | None:
        return value

    @classmethod
    def _parse(cls: Type[ListOfIds], value: list[int] | list[str]) -> list[int] | list[str] | None:  # type: ignore
        return value


@dataclass
class ListOfString:
    default: list[str] | None = None
    widget: str = "list_of_string"

    @classmethod
    def from_dict(cls: Type[ListOfString], **kwargs: Any) -> "ListOfString":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def _serialize(cls: Type[ListOfString], value: list[str] | None) -> list[str] | None:
        return value

    @classmethod
    def _parse(cls: Type[ListOfString], value: list[str]) -> list[str] | None:  # type: ignore
        return value


@dataclass
class RenameRule:
    group_name: str
    child_events: List[str]


@dataclass
class RenameRulesWidget:
    default: list[RenameRule] | None = None
    widget: str = "rename_rules"

    @classmethod
    def from_dict(cls: Type[RenameRulesWidget], **kwargs: Any) -> "RenameRulesWidget":
        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def _serialize(cls: Type[RenameRulesWidget], value: list[RenameRule] | None) -> list[RenameRule] | None:
        return value

    @classmethod
    def _parse(cls: Type[RenameRulesWidget], value: list[RenameRule] | None) -> list[RenameRule] | None:
        return value


WIDGET_TYPE = Union[
    Type[StringWidget],
    Type[IntegerWidget],
    Type[EnumWidget],
    Type[BooleanWidget],
    Type[ReteTimeWidget],
    Type[ReteFunction],
    Type[RenameRulesWidget],
]

WIDGET = Union[StringWidget, IntegerWidget, EnumWidget, BooleanWidget, ReteTimeWidget, ReteFunction, RenameRulesWidget]

# @TODO: make default dict. Vladimir Makhanov
WIDGET_MAPPING: dict[str, WIDGET_TYPE] = {
    "string": StringWidget,
    "integer": IntegerWidget,
    "enum": EnumWidget,
    "boolean": BooleanWidget,
    "tuple": ReteTimeWidget,
    "callable": ReteFunction,
}



from __future__ import annotations

from typing import Any, Dict

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import seaborn as sns

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
    tracker,
)
from retentioneering.eventstream.types import EventstreamType
from retentioneering.tooling.mixins.ended_events import EndedEventsMixin


class StepSankey(EndedEventsMixin):
    """
    A class for the visualization of user paths in stepwise manner using Sankey diagram.

    Parameters
    ----------
    eventstream : EventstreamType


    See Also
    --------
    .Eventstream.step_sankey : Call StepSankey tool as an eventstream method.
    .CollapseLoops : Find loops and create new synthetic events in the paths of all users having such sequences.
    .StepMatrix : This class provides methods for step matrix calculation and visualization.

    Notes
    -----
    See :doc:`StepSankey user guide</user_guides/step_sankey>` for the details.

    """

    max_steps: int
    threshold: int | float
    sorting: list | None
    targets: list[str] | str | None

    data_grp_nodes: pd.DataFrame
    data: pd.DataFrame
    data_grp_links: pd.DataFrame
    data_for_plot: dict

    @time_performance(
        scope="step_sankey",
        event_name="init",
    )
    def __init__(self, eventstream: EventstreamType) -> None:
        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp
        self.event_index_col = self.__eventstream.schema.event_index

        self.data_grp_nodes = pd.DataFrame()
        self.data = pd.DataFrame()
        self.data_grp_links = pd.DataFrame()
        self.data_for_plot = {}

    @staticmethod
    def _make_color(
        event: str,
        all_events: list,
        palette: list,
    ) -> tuple[int, int, int]:
        """
        It is a color picking function

        Parameters
        ----------
        event : str
            An event for color setting
        all_events : list
            A list of all events
        palette : list
            A list of colors

        Returns
        -------
        str
            A picked color for certain event
        """

        return palette[list(all_events).index(event)]

    @staticmethod
    def _round_up(
        n: float,
        dec: float,
    ) -> float:
        """
        Rounds the value up to the nearest value assuming a grid with ``dec`` step.
        E.g. ``_round_up(0.51, 0.05) = 0.55``, ``_round_up(0.55, 0.05) = 0.6``

        Parameters
        ----------
        n : float
            A number to round up
        dec : float
            A decimal for correct rounding up

        Returns
        -------
        float
            Rounded value
        """

        return round(n - n % dec + dec, 2)

    def _get_nodes_positions(self, df: pd.DataFrame) -> tuple[list[float], list[float]]:
        """
        It is a function for placing nodes at the x and y coordinates of plotly lib plot canvas.

        Parameters
        ----------
        df : pandas Dataframe
            A dataframe that contains aggregated information about the nodes.

        Returns
        -------
        tuple[list[float], list[float]]
            Two lists with the corresponding coordinates x and y.
        """
        # NOTE get x axis length
        x_len = len(df["step"].unique())

        # NOTE declare positions
        x_positions = []
        y_positions = []
        # NOTE get maximum range for placing middle points
        y_range = 0.95 - 0.05

        # NOTE going inside ranked events
        for step in sorted(df["step"].unique()):
            # NOTE placing x-axis points as well
            for _ in df[df["step"] == step][self.event_col]:
                x_positions.append([round(x, 2) for x in np.linspace(0.05, 0.95, x_len)][step - 1])

            # NOTE it always works very well if you have less than 4 values at current rank
            y_len = len(df[df["step"] == step][self.event_col])

            # NOTE at this case using came positions as x-axis because we don't need to calculate something more
            if y_len < 4:
                for p in [round(y, 2) for y in np.linspace(0.05, 0.95, y_len)]:
                    y_positions.append(p)

            # NOTE jumping in to complex part
            else:
                # NOTE total sum for understanding do we need extra step size or not
                total_sum = df[df["step"] == step]["usr_cnt"].sum()
                # NOTE step size for middle points
                y_step = round(y_range / total_sum, 2)
                # NOTE cumulative sum for understanding do we need use default step size or not
                cumulative_sum = 0
                # NOTE ENDED action
                ended_sum = df[(df["step"] == step) & (df[self.event_col] == "ENDED")]["usr_cnt"].sum()
                last_point = self._round_up(ended_sum / total_sum, 0.05)

                iterate_sum = 0

                # NOTE going deeper inside each event
                for n, event in enumerate(df[df["step"] == step][self.event_col]):
                    # NOTE placing first event at first possible position
                    if n == 0:
                        y_positions.append(0.05)

                    # NOTE placing last event at last possible position
                    elif n + 1 == y_len:
                        y_positions.append(0.95)

                    # NOTE placing middle points
                    else:
                        # NOTE we found out that 70% of total sum is the best cap for doing this case
                        if iterate_sum / total_sum > 0.2 and event != "ENDED":
                            # NOTE placing first point after the biggest one at the next position
                            # but inside [.1; .3] range
                            y_positions.append(
                                round(y_positions[-1] + np.minimum(np.maximum(y_step * iterate_sum, 0.1), 0.3), 2)
                            )

                        # NOTE placing points after the biggest
                        else:
                            # NOTE placing little points at the all available space
                            y_positions.append(
                                round(y_positions[-1] + (0.95 - last_point - y_positions[-1]) / (y_len - n), 2)
                            )

                    # NOTE set sum for next step
                    iterate_sum = df[(df["step"] == step) & (df[self.event_col] == event)]["usr_cnt"].to_numpy()[0]
                    # NOTE update cumulative sum
                    cumulative_sum += iterate_sum

        return x_positions, y_positions

    def _pad_end_events(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        If the number of events in a user's path is less than self.max_steps, then the function pads the path with
        multiply ENDED events. It is required for correct visualization of the trajectories which are
        shorter than self.max_steps.
        """
        pad = (
            data.groupby(self.user_col, as_index=False)[self.event_col]
            .count()
            .loc[lambda df_: df_[self.event_col] < self.max_steps]  # type: ignore
            .assign(repeat_number=lambda df_: self.max_steps - df_[self.event_col])
        )
        repeats = pd.DataFrame({self.user_col: np.repeat(pad[self.user_col], pad["repeat_number"])})
        padded_end_events = pd.merge(repeats, data[data[self.event_col] == "ENDED"], on=self.user_col)
        result = pd.concat([data, padded_end_events]).sort_values([self.user_col, self.event_index_col])
        return result

    def _prepare_data(self, data: pd.DataFrame) -> pd.DataFrame:
        data = self._add_ended_events(
            data=data, schema=self.__eventstream.schema, weight_col=self.__eventstream.schema.user_id
        )
        data = self._pad_end_events(data)
        # NOTE set new columns using declared functions
        data[self.time_col] = pd.to_datetime(data[self.time_col])
        data["step"] = data.groupby(self.user_col)[self.event_index_col].rank(method="first").astype(int)
        data = data.sort_values(by=["step", self.time_col]).reset_index(drop=True)
        data = self._get_next_event_and_timedelta(data)

        # NOTE threshold
        data["event_users"] = data.groupby(by=["step", self.event_col])[self.user_col].transform("nunique")
        data["total_users"] = data.loc[data["step"] == 1, self.user_col].nunique()
        data["perc"] = data["event_users"] / data["total_users"]

        if isinstance(self.threshold, float):
            column_to_compare = "perc"
        else:
            # assume that self.threshold must be of int type here
            column_to_compare = "event_users"

        events_to_keep = ["ENDED"]
        if self.targets is not None:
            events_to_keep += self.targets

        threshold_events = (
            data.loc[data["step"] <= self.max_steps, :]
            .groupby(by=self.event_col, as_index=False)[column_to_compare]
            .max()
            .loc[
                lambda df_: (df_[column_to_compare] <= self.threshold) & (~df_[self.event_col].isin(events_to_keep))
            ]  # type: ignore
            .loc[:, self.event_col]
        )
        data.loc[data[self.event_col].isin(threshold_events), self.event_col] = f"thresholded_{len(threshold_events)}"

        # NOTE rearrange the data taking into account recently added thresholded events
        data["step"] = data.groupby(self.user_col)[self.event_index_col].rank(method="first").astype(int)
        data = self._get_next_event_and_timedelta(data)

        # NOTE use max_steps for filtering data
        data = data.loc[data["step"] <= self.max_steps, :]

        # TODO: Do we really need to replace NA values?
        # NOTE skip mean calculating error
        data["time_to_next"] = data["time_to_next"].fillna(data["time_to_next"].min())
        return data

    def _render_plot(
        self,
        data_for_plot: dict,
        data_grp_nodes: pd.DataFrame,
        autosize: bool = True,
        width: int | None = None,
        height: int | None = None,
    ) -> go.Figure:
        # NOTE fill lists for plot
        targets = []
        sources = []
        values = []
        time_to_next = []
        for source_key in data_for_plot["links_dict"].keys():
            for target_key, target_value in data_for_plot["links_dict"][source_key].items():
                sources.append(source_key)
                targets.append(target_key)
                values.append(target_value["unique_users"])
                time_to_next.append(
                    str(pd.to_timedelta(target_value["avg_time_to_next"] / target_value["unique_users"])).split(".")[0]
                )
        # NOTE fill another lists for plot
        labels = []
        colors = []
        percs = []
        for key in data_for_plot["nodes_dict"].keys():
            labels += list(data_for_plot["nodes_dict"][key]["sources"])
            colors += list(data_for_plot["nodes_dict"][key]["color"])
            percs += list(data_for_plot["nodes_dict"][key]["percs"])
        # NOTE get colors for plot
        for idx, color in enumerate(colors):
            colors[idx] = "rgb" + str(color) + ""
        # NOTE get positions for plot
        x, y = self._get_nodes_positions(df=data_grp_nodes)
        # NOTE make plot
        fig = go.Figure(
            data=[
                go.Sankey(
                    arrangement="snap",
                    node=dict(
                        thickness=15,
                        line=dict(color="black", width=0.5),
                        label=labels,
                        color=colors,
                        customdata=percs,
                        hovertemplate="Total unique users: %{value} (%{customdata}% of total)<extra></extra>",
                        x=x,
                        y=y,
                        pad=20,
                    ),
                    link=dict(
                        source=sources,
                        target=targets,
                        value=values,
                        label=time_to_next,
                        hovertemplate="%{value} unique users went from %{source.label} to %{target.label}.<br />"
                        + "<br />It took them %{label} in average.<extra></extra>",
                    ),
                )
            ]
        )
        fig.update_layout(font=dict(size=15), plot_bgcolor="white", autosize=autosize, width=width, height=height)

        return fig

    def _get_links(
        self, data: pd.DataFrame, data_for_plot: dict, data_grp_nodes: pd.DataFrame
    ) -> tuple[dict, pd.DataFrame]:
        # NOTE create links aggregated dataframe
        data_grp_links = (
            data[data["step"] <= self.max_steps - 1]
            .groupby(by=["step", self.event_col, "next_event"])[[self.user_col, "time_to_next"]]
            .agg({self.user_col: ["count"], "time_to_next": ["sum"]})
            .reset_index()
            .rename(columns={self.user_col: "usr_cnt", "time_to_next": "time_to_next_sum"})
        )
        data_grp_links.columns = data_grp_links.columns.droplevel(1)
        data_grp_links = data_grp_links.merge(
            data_grp_nodes[["step", self.event_col, "index"]],
            how="inner",
            on=["step", self.event_col],
        )
        data_grp_links.loc[:, "next_step"] = data_grp_links["step"] + 1
        data_grp_links = data_grp_links.merge(
            data_grp_nodes[["step", self.event_col, "index"]].rename(
                columns={"step": "next_step", self.event_col: "next_event", "index": "next_index"}
            ),
            how="inner",
            on=["next_step", "next_event"],
        )

        data_grp_links.sort_values(by=["index", "usr_cnt"], ascending=[True, False], inplace=True)
        data_grp_links.reset_index(drop=True, inplace=True)

        # NOTE generating links plot dict
        data_for_plot.update({"links_dict": dict()})
        for index in data_grp_links["index"].unique():
            for next_index in data_grp_links[data_grp_links["index"] == index]["next_index"].unique():
                _unique_users, _avg_time_to_next = (
                    data_grp_links.loc[
                        (data_grp_links["index"] == index) & (data_grp_links["next_index"] == next_index),
                        ["usr_cnt", "time_to_next_sum"],
                    ]
                    .to_numpy()
                    .T
                )

                if index in data_for_plot["links_dict"]:
                    if next_index in data_for_plot["links_dict"][index]:
                        data_for_plot["links_dict"][index][next_index]["unique_users"] = _unique_users[0]
                        data_for_plot["links_dict"][index][next_index]["avg_time_to_next"] = np.timedelta64(
                            _avg_time_to_next[0]
                        )
                    else:
                        data_for_plot["links_dict"][index].update(
                            {
                                next_index: {
                                    "unique_users": _unique_users[0],
                                    "avg_time_to_next": np.timedelta64(_avg_time_to_next[0]),
                                }
                            }
                        )
                else:
                    data_for_plot["links_dict"].update(
                        {
                            index: {
                                next_index: {
                                    "unique_users": _unique_users[0],
                                    "avg_time_to_next": np.timedelta64(_avg_time_to_next[0]),
                                }
                            }
                        }
                    )
        return data_for_plot, data_grp_links

    def _get_nodes(self, data: pd.DataFrame) -> tuple[dict, pd.DataFrame]:
        all_events = list(data[self.event_col].unique())
        palette = self._prepare_palette(all_events)

        # NOTE create nodes aggregate dataframe
        data_grp_nodes = (
            data.groupby(by=["step", self.event_col])[self.user_col]
            .nunique()
            .reset_index()
            .rename(columns={self.user_col: "usr_cnt"})
        )
        data_grp_nodes.loc[:, "usr_cnt_total"] = data_grp_nodes.groupby(by=["step"])["usr_cnt"].transform("sum")
        data_grp_nodes.loc[:, "perc"] = np.round(
            (data_grp_nodes.loc[:, "usr_cnt"] / data_grp_nodes.loc[:, "usr_cnt_total"]) * 100, 2
        )
        data_grp_nodes.sort_values(
            by=["step", "usr_cnt", self.event_col],
            ascending=[True, False, True],
            inplace=True,
        )
        data_grp_nodes.reset_index(
            drop=True,
            inplace=True,
        )
        data_grp_nodes.loc[:, "color"] = data_grp_nodes[self.event_col].apply(
            lambda x: self._make_color(x, all_events, palette)
        )
        data_grp_nodes.loc[:, "index"] = data_grp_nodes.index  # type: ignore
        # NOTE doing right ranking
        if self.sorting is None:
            data_grp_nodes.loc[:, "sorting"] = 100
        else:
            for n, s in enumerate(self.sorting):
                data_grp_nodes.loc[data_grp_nodes[self.event_col] == s, "sorting"] = n
            data_grp_nodes.loc[:, "sorting"].fillna(100, inplace=True)
        # NOTE placing ENDED at the end
        data_grp_nodes.loc[data_grp_nodes[self.event_col] == "ENDED", "sorting"] = 101
        # NOTE using custom ordering
        data_grp_nodes.loc[:, "sorting"] = data_grp_nodes.loc[:, "sorting"].astype(int)

        # @TODO: step variable is not used inside the loop. The loop might be invalid. Vladimir Kukushkin
        # NOTE doing loop for valid ranking
        for step in data_grp_nodes["step"].unique():
            # NOTE saving last level order
            data_grp_nodes.loc[:, "order_by"] = (
                data_grp_nodes.groupby(by=[self.event_col])["index"].transform("shift").fillna(100).astype(int)
            )

            # NOTE placing ENDED events at the end
            data_grp_nodes.loc[data_grp_nodes[self.event_col] == "ENDED", "sorting"] = 101

            # NOTE creating new indexes
            data_grp_nodes.sort_values(
                by=["step", "sorting", "order_by", "usr_cnt", self.event_col],
                ascending=[True, True, True, False, True],
                inplace=True,
            )

            data_grp_nodes.reset_index(
                drop=True,
                inplace=True,
            )

            data_grp_nodes.loc[:, "index"] = data_grp_nodes.index  # type: ignore

        # NOTE generating nodes plot dict
        data_for_plot: Dict[str, Any] = dict()
        data_for_plot.update({"nodes_dict": dict()})
        for step in data_grp_nodes["step"].unique():
            data_for_plot["nodes_dict"].update({step: dict()})
            _sources, _color, _sources_index, _percs = (
                data_grp_nodes.loc[data_grp_nodes["step"] == step, [self.event_col, "color", "index", "perc"]]
                .to_numpy()
                .T
            )

            data_for_plot["nodes_dict"][step].update(
                {
                    "sources": list(_sources),
                    "color": list(_color),
                    "sources_index": list(_sources_index),
                    "percs": list(_percs),
                }
            )

        return data_for_plot, data_grp_nodes

    @staticmethod
    def _prepare_palette(all_events: list) -> list[tuple]:
        # NOTE default color palette
        palette_hex = ["50BE97", "E4655C", "FCC865", "BFD6DE", "3E5066", "353A3E", "E6E6E6"]
        # NOTE convert HEX to RGB
        palette = []
        for color in palette_hex:
            rgb_color = tuple(int(color[i : i + 2], 16) for i in (0, 2, 4))
            palette.append(rgb_color)

        # NOTE extend color palette if number of events more than default colors list
        complementary_palette = sns.color_palette("deep", len(all_events) - len(palette))
        if len(complementary_palette) > 0:
            colors = complementary_palette.as_hex()
            for c in colors:
                col = c[1:]
                palette.append(tuple(int(col[i : i + 2], 16) for i in (0, 2, 4)))

        return palette

    def _get_next_event_and_timedelta(self, data: pd.DataFrame) -> pd.DataFrame:
        grouped = data.groupby(self.user_col)
        data["next_event"] = grouped[self.event_col].shift(-1)
        data["next_timestamp"] = grouped[self.time_col].shift(-1)
        data["time_to_next"] = data["next_timestamp"] - data[self.time_col]
        data = data.drop("next_timestamp", axis=1)
        return data

    @time_performance(scope="step_sankey", event_name="fit")
    def fit(
        self,
        max_steps: int = 10,
        threshold: int | float = 0.05,
        sorting: list | None = None,
        targets: list[str] | str | None = None,
    ) -> None:
        """
        Calculate the sankey diagram internal values with the defined parameters.
        Applying ``fit`` method is necessary for the following usage
        of any visualization or descriptive ``StepSankey`` methods.

        Parameters
        ----------
        max_steps : int, default 10
            Maximum number of steps in trajectories to include. Should be > 1.
        threshold : float | int, default 0.05
            Used to remove rare events from the plot. An event is collapsed to ``thresholded_N`` artificial event if
            its maximum frequency across all the steps is less than or equal to ``threshold``. The frequency is set
            with respect to ``threshold`` type:

            - If ``int`` - the frequency is the number of unique users who had given event at given step.
            - If ``float`` - percentage of users: the same as for ``int``, but divided by the number of unique users.

            The events which are prohibited for collapsing could be enlisted in ``target`` parameter.
        sorting : list of str, optional
            Define the order of the events visualized at each step. The events that are not represented in the list
            will follow after the events from the list.
        targets : str or list of str, optional
            Contain events that are prohibited for collapsing with ``threshold`` parameter.

        Raises
        ------
        ValueError
            If ``max_steps`` parameter is <= 1.
        """
        data = self.__eventstream.to_dataframe(copy=True)[
            [self.user_col, self.event_col, self.time_col, self.event_index_col]
        ]
        if max_steps <= 1:
            raise ValueError("max_steps parameter must be > 1!")
        called_params = {
            "max_steps": max_steps,
            "threshold": threshold,
            "sorting": sorting,
            "targets": targets,
        }

        self.max_steps = max_steps
        self.threshold = threshold
        self.sorting = sorting
        self.targets = targets

        self.data = self._prepare_data(data)
        data_for_plot, self.data_grp_nodes = self._get_nodes(self.data)
        self.data_for_plot, self.data_grp_links = self._get_links(self.data, data_for_plot, self.data_grp_nodes)
        performance_info = {
            "nodes_count": len(self.data_grp_nodes),
            "links_count": len(self.data_grp_links),
        }
        collect_data_performance(
            scope="step_sankey",
            event_name="metadata",
            called_params=called_params,
            performance_data=performance_info,
            eventstream_index=self.__eventstream._eventstream_index,
        )

    @time_performance(
        scope="step_sankey",
        event_name="plot",
    )
    def plot(self, autosize: bool = True, width: int | None = None, height: int | None = None) -> go.Figure:
        """
        Create a Sankey interactive plot based on the calculated values.
        Should be used after :py:func:`fit`.

        Parameters
        ----------
        autosize : bool, default True
            Plotly autosize parameter. See :plotly_autosize:`plotly documentation<>`.
        width : int, optional
            Plot's width (in px). See :plotly_width:`plotly documentation<>`.
        height : int, optional
            Plot's height (in px). See :plotly_height:`plotly documentation<>`.

        Returns
        -------
        plotly.graph_objects.Figure

        """
        called_params = {
            "autosize": autosize,
            "width": width,
            "height": height,
        }

        figure = self._render_plot(
            data_for_plot=self.data_for_plot,
            data_grp_nodes=self.data_grp_nodes,
            autosize=autosize,
            width=width,
            height=height,
        )

        performance_info = {
            "nodes_count": len(self.data_grp_nodes),
            "links_count": len(self.data_grp_links),
        }
        collect_data_performance(
            scope="step_sankey",
            event_name="metadata",
            called_params=called_params,
            performance_data=performance_info,
            eventstream_index=self.__eventstream._eventstream_index,
        )

        return figure

    @property
    @time_performance(
        scope="step_sankey",
        event_name="values",
    )
    def values(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Returns two pd.DataFrames which the Sankey diagram is based on.

        Should be used after :py:func:`fit`.

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame]
            1. Contains the nodes of the diagram.
            2. Contains the edges of the diagram.

        """
        return self.data_grp_nodes, self.data_grp_links

    @property
    @time_performance(
        scope="step_sankey",
        event_name="params",
    )
    def params(self) -> dict:
        """
        Returns the parameters used for the last fitting.
        Should be used after :py:func:`fit`.

        """
        return {
            "max_steps": self.max_steps,
            "threshold": self.threshold,
            "sorting": self.sorting,
            "targets": self.targets,
        }



# type: ignore
# @FIXME: All code about subprocesses and uuids return either str or bytes or something else. Vladimir Makhanov.
import os
import platform
import subprocess
import uuid
from collections import defaultdict

UNDEFINED_PLATFORM_UUID = uuid.UUID("00000000-0000-0000-aaaa-eeeeeeeeeeee")
UNDEFINED_WINDOWS_PLATFORM_UUID = uuid.UUID("00000000-0000-0000-aaaa-aaaaaaaaaaaa")
UNDEFINED_LINUX_PLATFORM_UUID = uuid.UUID("00000000-0000-0000-aaaa-bbbbbbbbbbbb")
UNDEFINED_MAC_PLATFORM_UUID = uuid.UUID("00000000-0000-0000-aaaa-cccccccccccc")


def get_windows_hwid() -> str:
    try:
        current_machine_id = (
            str(subprocess.check_output("wmic csproduct get uuid", stderr=subprocess.DEVNULL), "utf-8")
            .split("\n")[1]
            .strip()
        )
        hwid = str(uuid.UUID(current_machine_id))
    except Exception:  # not sure if this is the right exception, but I don't know right one. Vladimir Makhanov.
        hwid = ""
    return hwid


def get_linux_hwid() -> str:
    try:
        if os.access("/etc/machine-id", os.R_OK):
            hw = subprocess.check_output(["cat", "/etc/machine-id"], stderr=subprocess.DEVNULL)
            hw = hw.decode().strip()
            hwid = str(uuid.UUID(hw))
        else:
            hwid = ""
    except Exception:  # not sure if this is the right exception, but I don't know right one. Vladimir Makhanov.
        hwid = ""
    return hwid


def undefined_platform() -> str:
    return ""


__platforms_HWID = defaultdict(default_factory=undefined_platform)
__platforms_HWID["Windows"] = get_windows_hwid
__platforms_HWID["Linux"] = get_linux_hwid


def get_hwid() -> str:
    try:
        os_name = platform.system()
        hwid = __platforms_HWID[os_name]()
    except Exception:
        hwid = ""
    return hwid


__all__ = ("get_hwid",)



from __future__ import annotations

from typing import Literal, Union

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class CollapseLoopsHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(
        scope="collapse_loops",
        event_name="helper",
        event_value="combine",
    )
    def collapse_loops(
        self: EventstreamType,
        suffix: Union[Literal["loop", "count"], None] = None,
        time_agg: Literal["max", "min", "mean"] = "min",
    ) -> EventstreamType:
        """
        A method of ``Eventstream`` class that finds ``loops`` and creates new synthetic events
        in paths of all users having such sequences.

        A ``loop`` - is a sequence of repetitive events.
        For example *"event1 -> event1"*

        Parameters
        ----------
            %(CollapseLoops.parameters)s

        Returns
        -------
        Eventstream
             Input ``eventstream`` with ``loops`` replaced by new synthetic events.


        """
        calling_params = {
            "suffix": suffix,
            "time_agg": time_agg,
        }
        not_hash_values = ["suffix", "time_agg"]

        # avoid circular import
        from retentioneering.data_processors_lib import (
            CollapseLoops,
            CollapseLoopsParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(
            processor=CollapseLoops(params=CollapseLoopsParams(suffix=suffix, time_agg=time_agg))  # type: ignore
        )
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="collapse_loops",
            event_name="metadata",
            called_params=calling_params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



# type: ignore

import random

import pandas as pd


def shuffle_df(df: pd.DataFrame, inplace=True):
    df = df if inplace else df.copy()
    index = [i for i in range(df.shape[0])]
    random.shuffle(index)
    df.set_index([index]).sort_index()
    return df


def get_merged_col(df: pd.DataFrame, colname: str, suffix: str):
    return df[colname] if colname in df else df[f"{colname}{suffix}"]


__all__ = ["shuffle_df", "get_merged_col"]



from inspect import signature
from typing import Callable, Optional

import pandas as pd
from pandas import DataFrame, Series

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.schema import EventstreamSchema
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ReteFunction


class PipeParams(ParamsModel):
    func: Callable[[DataFrame, Optional[EventstreamSchema]], DataFrame]

    _widgets = {
        "func": ReteFunction(),
    }


@docstrings.get_sections(base="Pipe")  # type: ignore
class Pipe(DataProcessor):
    """
    Modify an input eventstream in an arbitrary way by applying given function.
    The function must accept a DataFrame associated with the input eventstream
    and return a new state of the modified eventstream.

    Parameters
    ----------
    func : Callable[[DataFrame], DataFrame]
        A function that is applied to the DataFrame underlying the eventstream.
        Must accept DataFrame as input and return DataFrame as output

    Returns
    -------
    Eventstream
        Resulting eventstream
    """

    params: PipeParams

    @time_performance(
        scope="pipe",
        event_name="init",
    )
    def __init__(self, params: PipeParams):
        super().__init__(params=params)

    @time_performance(
        scope="pipe",
        event_name="apply",
    )
    def apply(self, df: DataFrame, schema: EventstreamSchemaType) -> DataFrame:
        func: Callable[[DataFrame, Optional[EventstreamSchemaType]], DataFrame] = self.params.func  # type: ignore

        expected_args_count = len(signature(func).parameters)
        if expected_args_count == 1:
            result = func(df)  # type: ignore
        else:
            result = func(df, schema)

        collect_data_performance(
            scope="pipe",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



import json
import traceback
from typing import Any, Optional, TypedDict

from .base import BaseReteException


class ServerNotFoundActionError(BaseReteException):
    message: str
    method: str

    def __init__(self, message: str, method: str):
        self.message = message
        self.method = method


class ServerNotFound(BaseReteException):
    pass


class ServerErrorDict(TypedDict):
    type: str
    msg: str
    errors: Optional[Any]
    formatted: str


class ServerErrorWithResponse(BaseReteException):
    type: str
    message: str
    errors: Optional[Any]
    formatted: str

    def __init__(self, message: str, type: str, errors: Any = None):
        self.message = message
        self.type = type
        self.errors = errors
        self.formatted = traceback.format_exc()

    def dict(self) -> ServerErrorDict:
        # check errors serializable
        message = self.message
        errors = self.errors

        try:
            json.dumps(self.errors)
        except:
            message = "serialize error response error!"
            errors = None

        return {
            "type": self.type,
            "msg": message,
            "errors": errors,
            "formatted": self.formatted,
        }



from __future__ import annotations

from typing import Any

from retentioneering.data_processor.registry import dataprocessor_view_registry

# import for filling registry
from retentioneering.data_processors_lib import AddPositiveEventsParams  # noqa


def list_dataprocessor(payload: dict[str, Any]) -> list[dict[str, str]]:
    registry: list[dict[str, str]] = dataprocessor_view_registry.get_export_registry()
    registry = sorted(registry, key=lambda x: x["name"])
    return registry



from __future__ import annotations

import pandas as pd

from retentioneering.eventstream.types import EventstreamType


class _DescribeEvents:
    TIME_ROUND_UNIT = "s"
    STATS_ORDER = ["mean", "std", "median", "min", "max"]
    UNIQUE_SESS = ("basic_statistics", "unique_sessions")
    SHARE_SESS = ("basic_statistics", "unique_sessions_shared")

    def __init__(
        self,
        eventstream: EventstreamType,
        session_col: str = "session_id",
        raw_events_only: bool = False,
        event_list: list[str] | None = None,
    ) -> None:
        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp
        self.session_col = session_col
        self.type_col = self.__eventstream.schema.event_type
        self.event_list = event_list
        self.raw_events_only = raw_events_only
        self.df = self.__eventstream.to_dataframe(copy=True)
        self.has_session_col: bool = False
        if self.raw_events_only:
            self.df = self.df[self.df[self.type_col].isin(["raw"])]

        self.total_events_base: int = len(self.df)
        self.unique_users_base: int = self.df[self.user_col].nunique()
        if self.session_col in self.df.columns:
            self.has_session_col = True
            self.total_sessions_base: int = self.df[self.session_col].nunique()

    def _agg_min_time(self, df: pd.DataFrame, agg_col: str, prefix: str) -> pd.DataFrame:
        df[f"__event_{prefix}_idx"] = df.groupby(agg_col).cumcount()
        df[f"__event_{prefix}_timedelta"] = df[self.time_col] - df.groupby(agg_col)[self.time_col].transform("first")
        return df

    def _agg_time_events(self, df: pd.DataFrame, agg_col: str, prefix: str) -> pd.DataFrame:
        first_time = f"time_to_FO_{prefix}_wise"
        first_event = f"steps_to_FO_{prefix}_wise"

        df_agg_event = (
            df.groupby([self.event_col, agg_col])
            .agg(time_to_FO=(f"__event_{prefix}_timedelta", "first"), steps_to_FO=(f"__event_{prefix}_idx", "first"))
            .reset_index()
        )
        df_agg_event.columns = [self.event_col, self.user_col, first_time, first_event]  # type: ignore
        df_agg_event[first_time] = df_agg_event[first_time].dt.total_seconds()
        df_agg_event = df_agg_event.groupby([self.event_col])[[first_time, first_event]].agg(
            self.STATS_ORDER  # type: ignore
        )
        df_agg_event[(first_event, "mean")] = df_agg_event[(first_event, "mean")].round(2)
        df_agg_event[(first_event, "std")] = df_agg_event[(first_event, "std")].round(2)

        for stat in self.STATS_ORDER:
            mult_ind = (first_time, stat)
            df_agg_event[mult_ind] = pd.to_timedelta(df_agg_event[mult_ind], unit="s").round(self.TIME_ROUND_UNIT)  # type: ignore

        return df_agg_event

    def _create_basic_info_df(self, df: pd.DataFrame) -> pd.DataFrame:
        basic_info = df.groupby("event").agg(
            number_of_occurrences=("event_id", "count"), unique_users=("user_id", "nunique")
        )

        basic_info["number_of_occurrences_shared"] = (
            basic_info["number_of_occurrences"] / self.total_events_base
        ).round(2)
        basic_info["unique_users_shared"] = (basic_info["unique_users"] / self.unique_users_base).round(2)

        basic_info.columns = pd.MultiIndex.from_product([["basic_statistics"], basic_info.columns])

        if self.has_session_col:
            basic_info[self.UNIQUE_SESS] = self.df.groupby("event")[self.session_col].agg("nunique")
            basic_info[self.SHARE_SESS] = (basic_info[self.UNIQUE_SESS] / self.total_sessions_base).round(2)

        return basic_info

    def _values(self) -> pd.DataFrame:
        df = self._agg_min_time(df=self.df, agg_col=self.user_col, prefix="user")

        if self.has_session_col:
            df = self._agg_min_time(df=df, agg_col=self.session_col, prefix="session")

        if self.event_list:
            df = df[df[self.event_col].isin(self.event_list)]

        basic_info = self._create_basic_info_df(df=df)
        df_agg_event = self._agg_time_events(df=df, agg_col=self.user_col, prefix="user")

        if self.has_session_col:
            df_agg_sess = self._agg_time_events(df=df, agg_col=self.session_col, prefix="session")
            df_agg_event = df_agg_event.merge(df_agg_sess, left_index=True, right_index=True)

        res = basic_info.merge(df_agg_event, left_index=True, right_index=True)
        if self.has_session_col:
            res.insert(2, self.UNIQUE_SESS, res.pop(self.UNIQUE_SESS))  # type: ignore
        return res


__all__ = ("_DescribeEvents",)



import base64
import io
import urllib
from typing import Callable, Optional

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from scipy.stats import wasserstein_distance

from retentioneering.common.formatters import truncate_str
from retentioneering.eventstream.segments import _split_segment
from retentioneering.eventstream.types import EventstreamType, SplitExpr


class SegmentDiff:
    def __init__(self, eventstream: EventstreamType):
        self.__eventstream = eventstream

    def fit(
        self,
        features: pd.DataFrame,
        segment_items: SplitExpr,
        aggfunc: Callable = np.mean,
        threshold: float = 0.01,
        top_n: Optional[int] = None,
    ) -> pd.DataFrame:
        from retentioneering.eventstream.eventstream import Eventstream

        stream: Eventstream = self.__eventstream  # type: ignore
        segment_name, segment_value_1, segment_value_2 = segment_items
        stream.materialize_segment(segment_name)

        user_groups, user_group_names = _split_segment(stream, segment_items)

        segment_value_1_users, segment_value_2_users = user_groups
        features_1 = features.loc[segment_value_1_users].assign(segment_value=segment_value_1)
        features_2 = features.loc[segment_value_2_users].assign(segment_value=segment_value_2)

        if len(features_1) < 10:
            raise ValueError(
                f"Segment {segment_name}={segment_value_1} contains less than 10 paths."
                "The density diagrams will be meaningless."
            )

        if len(features_2) < 10:
            raise ValueError(
                f"Segment {segment_name}={segment_value_2} contains less than 10 paths."
                "The density diagrams will be meaningless"
            )

        segment_name_1 = truncate_str(segment_value_1, max_len=20)
        segment_name_2 = truncate_str(segment_value_2, max_len=20)

        figsize = (4, 0.7)
        right_adjustment = max(0.5, 0.7 - (max(len(segment_name_1), len(segment_name_2)) / 20) * 0.3)

        res_list = []
        for feature in features.columns:
            f1 = features_1[feature]
            f2 = features_2[feature]
            f1_agg = f1.pipe(aggfunc)
            f2_agg = f2.pipe(aggfunc)
            wd = wasserstein_distance(f1, f2)

            if wd > threshold:
                fig, ax = plt.subplots(figsize=figsize)

                if f1.nunique() > 1:
                    sns.kdeplot(x=f1, bw_method=0.3, ax=ax, label=segment_name_1, color="C0", warn_singular=False)
                    ax.axvline(x=f1_agg, linestyle="--", color="C0")
                else:
                    sns.barplot(x=f1, ax=ax)
                if f2.nunique() > 1:
                    sns.kdeplot(x=f2, bw_method=0.3, ax=ax, label=segment_name_2, color="C1", warn_singular=False)
                    ax.axvline(x=f2_agg, linestyle="--", color="C1")
                else:
                    sns.barplot(x=f2, ax=ax)

                ax.legend(loc="center left", bbox_to_anchor=(1.05, 0.5))

                ax.yaxis.set_visible(False)
                ax.set_xlabel("")
                ax.margins(x=0, y=0)
                plt.subplots_adjust(left=0, right=right_adjustment, top=1, bottom=0.3)

                # convert plt output to base64 and insert it as <img> html tag.
                buf = io.BytesIO()
                plt.savefig(buf, format="png")
                plt.close()
                buf.seek(0)
                img_string = base64.b64encode(buf.read())
                img_uri = "data:image/png;base64," + urllib.parse.quote(img_string)  # type: ignore
                img_html = '<img src = "%s"/>' % img_uri
            else:
                img_html = ""

            res_list.append([feature, f1_agg, f2_agg, wd, img_html])

        mertic_1_col = f"{segment_name}={segment_name_1}"
        mertic_2_col = f"{segment_name}={segment_name_2}"

        res = (
            pd.DataFrame(res_list, columns=["feature", mertic_1_col, mertic_2_col, "distance", "feature distribution"])
            .sort_values("distance", ascending=False)
            .reset_index(drop=True)
        )

        if top_n is not None:
            res = res.head(top_n)
        else:
            res = res[res["distance"] > threshold]

        return res



from __future__ import annotations

from typing import List

import pandas as pd
from pydantic.dataclasses import dataclass

from retentioneering.backend.tracker import track
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.widget.widgets import RenameRulesWidget


@dataclass
class RenameRule:
    group_name: str
    child_events: List[str]


class RenameParams(ParamsModel):
    """
    [{
        "group_name": "some_group",
        "child_events": ["eventA", "eventB"]
    }]
    """

    rules: List[RenameRule]

    _widgets = {
        "rules": RenameRulesWidget(),
    }


class RenameProcessor(DataProcessor):
    params: RenameParams

    def __init__(self, params: RenameParams):
        super().__init__(params=params)

    @track(  # type: ignore
        tracking_info={"event_name": "apply"},
        scope="rename",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        event_col = schema.event_name

        rename_rules: dict[str, str] = dict()
        for rule in self.params.rules:
            to_ = rule.group_name
            for from_ in rule.child_events:
                rename_rules[from_] = to_

        df[event_col] = df[event_col].replace(rename_rules)
        return df



import json
import os
import threading
from dataclasses import asdict

import pandas as pd

from ..tracking_info import TrackingInfo
from .protocol import ConnectorProtocol


class CSVConnector(ConnectorProtocol):
    def __init__(self, csv_name: str):
        super().__init__()
        # validate csv path writable

        csv_path = f"{os.getcwd()}/{csv_name}"
        if os.access(csv_path, os.F_OK):
            raise Exception(f"CSV file {csv_path} already exists")
        else:
            self.csv_path = csv_path
            os.environ["RETE_TRACKER_CSV_PATH"] = csv_path

        self.source = "rete_tools_backend"
        self.columns = [
            "user_id",
            "event_custom_name",
            "event_name",
            "event_value",
            "params",
            "scope",
            "event_time",
            "jupyter_kernel_id",
            "colab",
            "event_date_local",
            "event_day_week",
            "event_timestamp",
            "event_timestamp_ms",
            "source",
            "version",
            "os",
            "index",
            "browser",
            "eventstream_index",
            "parent_eventstream_index",
            "child_eventstream_index",
            "account_id",
        ]

    def _post(self, data: dict) -> None:
        # run _post_job in a separate thread using multi-threading
        try:
            post_thread = threading.Thread(target=self._post_job, args=(data,))
            post_thread.start()
        except Exception as e:
            pass

    def _post_job(self, data: dict) -> None:
        with open(self.csv_path, "a") as log_file:
            pd.DataFrame([data], columns=self.columns).to_csv(log_file, index=False, header=False)

    def _prepare_data(self, data: TrackingInfo) -> dict:
        prepared_data = asdict(data)
        prepared_data["source"] = self.source
        prepared_data["params"] = json.dumps(prepared_data["params"])
        del prepared_data["event_time"]
        return prepared_data

    def send_message(self, data: TrackingInfo) -> None:
        prepared_data = self._prepare_data(data=data)
        self._post(prepared_data)



from __future__ import annotations

from typing import Any, Literal, Optional, Tuple, Union

import matplotlib.axes
import networkx as nx
import pandas as pd
import seaborn as sns

from retentioneering.edgelist import Edgelist
from retentioneering.eventstream.helpers import FilterEventsHelperMixin
from retentioneering.eventstream.segments import _split_segment
from retentioneering.eventstream.types import (
    EventstreamType,
    SplitExpr,
    UserGroupsNamesType,
    UserGroupsType,
)
from retentioneering.nodelist import Nodelist
from retentioneering.tooling.transition_graph.types import NormType

MAX_DIM = 60
SHOW_VALUES_DIM = 30
SEQUENCES_URL = "https://doc.retentioneering.com/stable/doc/user_guides/sequences.html"
TRANSITION_MATRIX_VALUES_URL = "https://doc.retentioneering.com/stable/doc/user_guides/transition_matrix.html#values"


class TransitionMatrix:
    """
    The TransitionMatrix class represents a matrix where the element at position (i, j)
    displays the weight of the transition from event i to event j.
    This class provides methods for calculating and visualizing transition matrices,
    using the same logic as for calculating edge weights in a transition graph.

    Parameters
    ----------
    eventstream : EventstreamType
        The eventstream for which the transition matrix is computed.

    See Also
    --------
    .Eventstream.transition_matrix : This method can be called on an Eventstream to obtain a TransitionMatrix.
    .TransitionGraph : An interactive tool for representing transitions as a graph.

    Notes
    -----
    For more detailed information, refer to the :doc:`Transition matrix user guide</user_guides/transition_matrix>`.
    """

    __eventstream: EventstreamType
    groups: UserGroupsType | None = None
    group_names: UserGroupsNamesType | None = None

    __nodelist: Nodelist
    __edgelist: Edgelist
    __values: pd.DataFrame
    __weight_col: str | None
    __norm_type: NormType
    __fill_value: Any
    __title: str

    def __init__(self, eventstream: EventstreamType) -> None:
        self._eventstream = eventstream
        self._nodelist = Nodelist(
            weight_cols=[eventstream.schema.event_id, *eventstream.schema.custom_cols],
            time_col=eventstream.schema.event_timestamp,
            event_col=eventstream.schema.event_name,
        )
        self._nodelist.calculate_nodelist(self._eventstream.to_dataframe())
        self._edgelist = Edgelist(eventstream=eventstream)
        self.__fill_value = 0
        self.__title = ""
        self.__norm_type = None
        self.__weight_col = None

    def fit(
        self,
        weight_col: Optional[str] = None,
        norm_type: NormType = None,
        groups: SplitExpr | None = None,
    ) -> None:
        """
        Calculates transition weights as a matrix for each unique pair of events.
        The calculation logic is the same that is used for edge weights calculation of transition graph.
        Applying ``fit`` method is necessary for the following usage
        of any visualization or descriptive ``TransitionMatrix`` methods.

        Parameters
        ----------
        norm_type : {"full", "node", None}, default None
            Type of normalization that is used to calculate weights.
            Based on ``weight_col`` parameter the weight values are calculated.

            - If ``None``, normalization is not used, the absolute values are taken.
            - If ``full``, normalization across the whole eventstream.
            - If ``node``, normalization across each node (or outgoing transitions from each node).

            See :ref:`Transition graph user guide <transition_graph_weights>` for the details.

        weight_col : str, optional
            A column name from the :py:class:`.EventstreamSchema` which values will control the final
            edges' weights.

            For each edge is calculated:

            - If ``None`` or ``user_id`` - the number of unique users.
            - If ``event_id`` - the number of transitions.
            - If ``session_id`` - the number of unique sessions.
            - If ``custom_col`` - the number of unique values in selected column.

            See :ref:`Transition graph user guide <transition_graph_weights>` for the details.

        groups : tuple[list, list], tuple[str, str, str], str, optional
            Specify two groups of paths to plot differential transition matrix.
            Two transition matrices M1 and M2 will be calculated for these groups.
            Resulting matrix is M = M1 - M2.

            - If ``tuple[list, list]``, each sub-list should contain valid path ids.
            - If ``tuple[str, str, str]``, the first str should refer to a segment name,
              the others should refer to the corresponding segment values.
            - If ``str``, it should refer to a binary (i.e. containing two segment values only) segment name.

        """
        self.__weight_col = weight_col or self._eventstream.schema.user_id
        self.__norm_type = norm_type

        if groups:
            groups_, group_names = _split_segment(self._eventstream, groups)
            self.groups = groups_
            self.group_names = group_names

        if self.groups is None:
            self.__values = self._values(self.__weight_col, self.__norm_type)
            self.__title = "Transition matrix"
        else:
            event_list = list(self._nodelist.nodelist_df[self._nodelist.event_col])
            with pd.option_context("future.no_silent_downcasting", True):
                frame = pd.DataFrame(columns=event_list, index=event_list).fillna(self.__fill_value).infer_objects()
            positive_matrix = self._filter_group(self.groups[0])
            negative_matrix = self._filter_group(self.groups[1])
            self.__values = (
                frame.add(positive_matrix.__values, fill_value=self.__fill_value)
                .sub(negative_matrix.__values, fill_value=self.__fill_value)
                .fillna(self.__fill_value)
            )
            if self.groups:
                if self.group_names:
                    groups_subtitle = f", {self.group_names[0]} vs. {self.group_names[1]}"
                else:
                    groups_subtitle = ", group 1 vs. group 2"
            else:
                groups_subtitle = ""
            self.__title = (
                f"Differential transition matrix{groups_subtitle}\n"
                f"(group sizes: {positive_matrix._n_users}, {negative_matrix._n_users})"
            )

    def _filter_group(self, group: list) -> TransitionMatrix:
        if not isinstance(self._eventstream, FilterEventsHelperMixin):
            raise TypeError("filter_events is not implemented for the eventstream")
        substream = self._eventstream.filter_events(
            lambda df, schema: df[schema.user_id].isin(group)  # pyright: ignore [reportOptionalMemberAccess]
        )
        matrix = TransitionMatrix(substream)
        matrix.fit(weight_col=self.__weight_col, norm_type=self.__norm_type)
        return matrix

    def _values(self, weight_col: str, norm_type: NormType) -> pd.DataFrame:
        self._edgelist.calculate_edgelist(norm_type=norm_type, weight_cols=[weight_col])
        edgelist: pd.DataFrame = self._edgelist.edgelist_df.copy()
        edgelist = edgelist.drop(columns=["rete_is_out_of_threshold"])
        graph = nx.DiGraph()
        graph.add_weighted_edges_from(edgelist.values)

        return nx.to_pandas_adjacency(G=graph)

    @property
    def values(self) -> pd.DataFrame:
        """
        Returns the calculated transition matrix as a pandas.DataFrame.
        Should be used after :py:func:`fit`.
        """
        return self.__values.copy()

    @property
    def _n_users(self) -> int:
        return self._eventstream.to_dataframe()[self._eventstream.schema.user_id].nunique()

    def plot(
        self,
        heatmap_axis: Union[Literal["rows", "columns", "both"], int] = "both",
        precision: Union[int, Literal["auto"]] = "auto",
        figsize: Optional[Tuple[Union[float, int], Union[float, int]]] = None,
        show_large_matrix: Optional[bool] = None,
        show_values: Optional[bool] = None,
    ) -> Optional[matplotlib.axes.Axes]:
        """
        Create a heatmap plot based on the calculated transition matrix values.
        This method should be used after calling :py:func:`fit`.

        Parameters
        ----------
        heatmap_axis : {0 or 'rows', 1 or 'columns', 'both'}, default 'both'
            The axis for which the heatmap is to be generated.
            If specified, the heatmap will be created separately for the selected axis.
            If ``heatmap_axis='both'``, the heatmap will be applied to the entire matrix.

        figsize : tuple[float, float], default None
            The size of the visualization. The default size is calculated automatically depending
            on the matrix dimension and `precision` and `show_values` options.

        precision : int or str, default 'auto'
            The number of decimal digits to display after zero as fractions in the heatmap.
            If precision is ``auto``, the value will depend on the ``norm_type``:
            0 for ``norm_type=None``, and 2 otherwise.

        show_large_matrix : bool, optional
            If ``None`` the matrix is displayed only in case the matrix dimension <= 60.
            If ``True``, the matrix is plotted explicitly.

        show_values : bool, optional
            If ``None`` the matrix values are not displayed only in case the matrix dimension lies between 30 and 60.
            If ``True``, the matrix values are shown explicitly.
            If ``False``, the values are hidden, ``precision`` parameter is ignored in this case.

        Returns
        -------
        matplotlib.axes.Axes
            The Axes object containing the heatmap plot.

        """
        dim = self.__values.shape[0]

        if dim > MAX_DIM and not show_large_matrix:
            output = (
                f"The transition matrix has more than {MAX_DIM} events. We don't recommend to plot such large matrices,"
                + f" show_large_matrix=True or use the Sequences tool instead:\n"
                + f"{SEQUENCES_URL}\n"
                + "You can still get the matrix values as a pandas.DataFrame by retrieving the .values property:\n"
                + f"{TRANSITION_MATRIX_VALUES_URL}\n"
            )
            print(output)
            return None
        elif dim > SHOW_VALUES_DIM or show_values is False:
            dim_mode = "medium"
            annot = False
            linewidths = 0.01
            cbar = True
        else:
            dim_mode = "small"
            annot = self.__values
            linewidths = False
            cbar = False

        if precision == "auto":
            fmt = ".0f" if self.__norm_type is None else ".2f"
        # else precision is int
        else:
            fmt = f".{precision}f"

        if not figsize:
            # dim_mode == "medium"
            cell_size = 0.25
            if dim_mode == "small":
                cell_size = len(f"{self.__values.max().max():{fmt}}") * 0.05 + 0.4

            figsize = (round(self.values.shape[0] * cell_size), round(self.values.shape[0] * cell_size))

        grid_specs = {"wspace": 0.08, "hspace": 0.08}

        matrix = self._normalize(heatmap_axis)

        figure, axs = sns.mpl.pyplot.subplots(
            figsize=figsize,
            gridspec_kw=grid_specs,
        )

        heatmap = sns.heatmap(
            matrix, annot=annot, fmt=fmt, cmap="RdGy", center=0, cbar=cbar, linewidths=linewidths, linecolor="gray"
        )
        heatmap.set_title(self.__title, fontsize=16)

        sns.mpl.pyplot.sca(axs)
        sns.mpl.pyplot.yticks(rotation=0)
        return axs

    def _normalize(self, axis: Union[int, str]) -> pd.DataFrame:
        matrix = self.__values.copy()
        if axis == "both":
            return matrix
        if axis == 0 or axis == "rows":
            return matrix.div(matrix.abs().max(axis=1), axis=0).fillna(self.__fill_value)
        elif axis == 1 or axis == "columns":
            return matrix.div(matrix.abs().max(axis=0), axis=1).fillna(self.__fill_value)
        else:
            raise ValueError(f"no axis named {axis} for the transition matrix")


__all__ = ("TransitionMatrix",)



from typing import Union

import pandas as pd


def fancy_round(df: pd.DataFrame) -> pd.DataFrame:
    """
    Format DataFrame numbers to a fancy str format.
    """

    def round_number(num: Union[float, int]) -> str:
        if isinstance(num, float):
            if abs(num) < 10:
                num = round(num, 3)
            elif abs(num) < 100:
                num = round(num, 2)
            else:
                num = round(num, 1)

        if isinstance(num, pd.Timedelta):
            seconds = num.total_seconds()  # type: ignore
            if abs(seconds) > 86400:
                return f"{round((seconds / 86400), 1)}d"
            elif abs(seconds) > 3600:
                return f"{round((seconds / 3600), 1)}h"
            elif abs(seconds) > 60:
                return f"{round((seconds / 60), 1)}m"
            elif abs(seconds) > 1:
                return f"{round(seconds, 1)}s"
            elif seconds == 0:
                return "0"
            else:
                return f"{round(seconds / 0.001, 1)}ms"

        if abs(num) >= 1000000000:
            return f"{round(num / 1000000000)}B"
        elif abs(num) >= 1000000:
            return f"{round(num / 1000000)}M"
        elif abs(num) >= 1000:
            return f"{round(num / 1000)}K"
        else:
            return str(num)

    rounded_df = df.map(round_number)  # type:ignore
    return rounded_df


def truncate_str(line: str, max_len: int) -> str:
    # a stupid workaround to include segment names that start with _ to the matplotlib legend
    # https://github.com/matplotlib/matplotlib/issues/21295
    line = " " + line

    if len(line) > max_len:
        return line[: max_len - 3] + "..."
    else:
        return line



from __future__ import annotations

import warnings
from typing import Any, Literal

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.eventstream.types import EventstreamType
from retentioneering.tooling.constants import BINS_ESTIMATORS


class EventTimestampHist:
    """
    Plot the distribution of events over time. Can be useful for detecting
    time-based anomalies, and visualising general timespan of the eventstream.

    Parameters
    ----------
    eventstream : EventstreamType

    See Also
    --------
    .TimedeltaHist : Plot the distribution of the time deltas between two events.
    .UserLifetimeHist : Plot the distribution of user lifetimes.
    .Eventstream.describe : Show general eventstream statistics.
    .Eventstream.describe_events : Show general eventstream events statistics.

    Notes
    -----
    See :ref:`Eventstream user guide<eventstream_events_timestamp>` for the details.
    """

    __eventstream: EventstreamType
    raw_events_only: bool
    event_list: list[str] | None
    lower_cutoff_quantile: float | None
    upper_cutoff_quantile: float | None
    bins: int | Literal[BINS_ESTIMATORS]
    bins_to_show: np.ndarray
    values_to_plot: np.ndarray

    @time_performance(
        scope="event_timestamp_hist",
        event_name="init",
    )
    def __init__(self, eventstream: EventstreamType) -> None:
        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp

        self.bins_to_show = np.array([])
        self.values_to_plot = np.array([])

    def _remove_cutoff_values(self, series: pd.Series) -> pd.Series:
        idx = [True] * len(series)
        if self.upper_cutoff_quantile is not None:
            idx &= series <= series.quantile(self.upper_cutoff_quantile)
        if self.lower_cutoff_quantile is not None:
            idx &= series >= series.quantile(self.lower_cutoff_quantile)
        return series[idx]

    def __validate_input(
        self,
        lower_cutoff_quantile: float | None = None,
        upper_cutoff_quantile: float | None = None,
    ) -> tuple[float | None, float | None]:
        if lower_cutoff_quantile is not None:
            if not 0 < lower_cutoff_quantile < 1:
                raise ValueError("lower_cutoff_quantile should be a fraction between 0 and 1.")

        if upper_cutoff_quantile is not None:
            if not 0 < upper_cutoff_quantile < 1:
                raise ValueError("upper_cutoff_quantile should be a fraction between 0 and 1.")

        if lower_cutoff_quantile is not None and upper_cutoff_quantile is not None:
            if lower_cutoff_quantile > upper_cutoff_quantile:
                warnings.warn("lower_cutoff_quantile exceeds upper_cutoff_quantile; no data passed to the histogram")

        return upper_cutoff_quantile, lower_cutoff_quantile

    @time_performance(
        scope="event_timestamp_hist",
        event_name="fit",
    )
    def fit(
        self,
        raw_events_only: bool = False,
        event_list: list[str] | None = None,
        lower_cutoff_quantile: float | None = None,
        upper_cutoff_quantile: float | None = None,
        bins: int | Literal[BINS_ESTIMATORS] = 20,
    ) -> None:
        """
        Calculate values for the histplot.

        Parameters
        ----------
        raw_events_only : bool, default False
            If ``True`` - statistics will only be shown for raw events.
            If ``False`` - statistics will be shown for all events presented in your data.
        event_list : list of str, optional
            Specify events to be displayed.
        lower_cutoff_quantile : float, optional
            Specify time distance quantile as the lower boundary. The values below the boundary are truncated.
        upper_cutoff_quantile : float, optional
            Specify time distance quantile as the upper boundary. The values above the boundary are truncated.
        bins : int or str, default 20
            Generic bin parameter that can be the name of a reference rule or
            the number of bins. Passed to :numpy_bins_link:`numpy.histogram_bin_edges<>`.

        Returns
        -------
        None

        """
        called_params = {
            "raw_events_only": raw_events_only,
            "event_list": event_list,
            "lower_cutoff_quantile": lower_cutoff_quantile,
            "upper_cutoff_quantile": upper_cutoff_quantile,
            "bins": bins,
        }

        self.upper_cutoff_quantile, self.lower_cutoff_quantile = self.__validate_input(
            lower_cutoff_quantile,
            upper_cutoff_quantile,
        )

        self.event_list = event_list
        self.raw_events_only = raw_events_only
        self.bins = bins

        data = self.__eventstream.to_dataframe(copy=True)

        if self.raw_events_only:
            data = data[data["event_type"].isin(["raw"])]
        if self.event_list:
            data = data[data[self.event_col].isin(self.event_list)]

        values_to_plot = data[self.time_col]
        if self._remove_cutoff_values:  # type: ignore
            values_to_plot = self._remove_cutoff_values(values_to_plot).to_numpy()

        bins_to_show = np.histogram_bin_edges(pd.to_numeric(values_to_plot), bins=self.bins)
        bins_to_show = pd.to_datetime(bins_to_show).round("s")
        if len(values_to_plot) == 0:
            bins_to_show = np.array([])

        self.bins_to_show = bins_to_show  # type: ignore
        self.values_to_plot = values_to_plot  # type: ignore

        collect_data_performance(
            scope="event_timestamp_hist",
            event_name="metadata",
            called_params=called_params,
            performance_data={},
            eventstream_index=self.__eventstream._eventstream_index,
        )

    @property
    @time_performance(  # type: ignore
        scope="event_timestamp_hist",
        event_name="values",
    )
    def values(self) -> tuple[np.ndarray, np.ndarray]:
        """

        Returns
        -------
        tuple(np.ndarray, np.ndarray)

            1. The first array contains the values for histogram.
            2. The first array contains the bin edges.

        """
        return self.values_to_plot, self.bins_to_show

    @time_performance(
        scope="event_timestamp_hist",
        event_name="plot",
    )
    def plot(self, width: float = 6.0, height: float = 4.5) -> Any:
        """
        Create a sns.histplot based on the calculated values.

        Parameters
        ----------
        width : float, default 6.0
            Width in inches.
        height : float, default 4.5
            Height in inches.

        Returns
        -------
        :matplotlib_axes:`matplotlib.axes.Axes<>`
            The matplotlib axes containing the plot.

        """

        figsize = (width, height)
        plt.figure(figsize=figsize)

        hist = sns.histplot(self.values_to_plot, bins=self.bins)
        hist.set_title("Event timestamp histogram")

        collect_data_performance(
            scope="event_timestamp_hist",
            event_name="metadata",
            called_params={"width": width, "height": height},
            performance_data={},
            eventstream_index=self.__eventstream._eventstream_index,
        )

        return hist
